# coding: utf-8
from __future__ import annotations


import logging
import os
import socket
import time

from ewokscore.task import Task as EwoksTask
from nxtomomill import converter as nxtomomill_converter
from silx.io.utils import h5py_read_dataset
from silx.io.utils import open as open_hdf5

import tomwer.version
from tomwer.core import settings
from tomwer.core.process.task import BaseProcessInfo
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.nxtomoscan import NXtomoScan

from .rpcserver import DataListenerThread

try:
    # new HDF5Config class
    from nxtomomill.io.config import TomoHDF5Config as HDF5Config
except ImportError:
    from nxtomomill.io.config import HDF5Config

_logger = logging.getLogger(__name__)


def nxtomomill_input_callback(entry, desc):
    """note: for the data listener we want to avoid any user interaction.
    In order to avoid any time lost"""
    if entry == "energy":
        default_energy = 19.0
        _logger.warning(f"Energy is missing. Set it to default to {default_energy}")
        return default_energy
    else:
        _logger.warning(f"missing {entry}. Won't be set")
        return None


class _DataListenerTaskPlaceHolder(EwoksTask, output_names=["data"]):
    pass


class DataListener(BaseProcessInfo):
    # TODO: update this, DataListener is not properly a task
    """
    class able to connect with a redis database.
    During an acquisition at esrf the redis database expose the scan under
    acquisition. This allow use to know when a scan is finished.

    In order to insure connection the machine need to know where is located
    the 'bacon server'
    (see https://bliss.gitlab-pages.esrf.fr/bliss/master/bliss_data_life.html).
    This is why the 'BEACON_HOST' environment variable should be defined.

    For example for id19 - lbs 191 we have:
    export BEACON_HOST="europa"
    On id19 - europa we would have
    export BEACON_HOST="localhost"
    """

    TIMOUT_READ_FILE = 30
    "When the event 'scan_ended' is received all data might not have been write" " yet"

    SWMR_MODE = None
    """The bliss writer is not using the swmr mode. This class has independent behavior regarding the tomoscan / nxotmo get_swmr_mode which is
    dedicated to the internal tomotools behavior
    """

    def __init__(self):
        super().__init__()
        self._host = settings.JSON_RPC_HOST
        # if host is None then use hostname
        if self._host is None:
            self._host = socket.gethostname()
        self._port = settings.JSON_RPC_PORT
        self._listening_thread = None

    @staticmethod
    def program_name():
        """Name of the program used for this processing"""
        return "scan listener"

    @staticmethod
    def program_version():
        """version of the program used for this processing"""
        return tomwer.version.version

    @staticmethod
    def definition():
        """definition of the process"""
        "convert bliss sequence with a proposal to a NXTomo nexus file"

    @property
    def port(self):
        return self._port

    @property
    def host(self):
        return self._host

    def get_listening_thread(self):
        """
        return the listening thread to ear at tango server
        """
        if self._listening_thread is None:
            self._listening_thread = self.create_listening_thread()
        return self._listening_thread

    def create_listening_thread(self):
        """
        Procedure to create the listening thread
        :return: listening thread
        """
        return DataListenerThread(host=self.host, port=self.port)

    def delete_listening_thread(self):
        """
        Procedure to delete the listening thread
        """
        if hasattr(self, "_listening_thread"):
            if self._listening_thread is not None:
                self._listening_thread.stop()
                self._listening_thread.join(5)
            del self._listening_thread
        self._listening_thread = None

    def set_configuration(self, properties):
        # for now the NXProcess cannot be tune
        if isinstance(properties, HDF5Config):
            self._settings = properties.to_dict()
        elif isinstance(properties, dict):
            self._settings = properties
        else:
            raise TypeError(f"invalid type: {type(properties)}")

    def run(self):
        pass

    def process_sample_file(
        self,
        sample_file: str,
        entry: str,
        proposal_file: str,
        master_sample_file: str | None,
    ) -> tuple:
        """

        :param sample_file: file to be converted
        :param entry: entry in the tango .h5 file to be converter
        :return: tuple scan succeeded.
        """
        if not os.path.isfile(sample_file):
            raise ValueError(f"Given file {sample_file} does not exists.")
        scans = []
        for mf_emf in self.convert(bliss_file=sample_file, entry=entry):
            master_file, entry_master_file = mf_emf
            if master_file is not None and entry_master_file is not None:
                scan = NXtomoScan(scan=master_file, entry=entry_master_file)

                try:
                    # register process.
                    self.register_process(
                        process_file=scan.process_file,
                        entry=scan.entry,
                        results={"output_file": scan.master_file, "entry": scan.entry},
                        configuration={
                            "sample_file": sample_file,
                            "entry": entry,
                            "file_proposal": proposal_file,
                            "master_sample_file": master_sample_file,
                        },
                        process_index=scan.pop_process_index(),
                        overwrite=True,
                    )
                except Exception as e:
                    _logger.error(e)

                self._signal_scan_ready(scan)
                scans.append(scan)
        return tuple(scans)

    @staticmethod
    def get_proposal_file(process_file, entry):
        """Return the proposal file of the experimentation if registered by the
        data listener"""
        if entry is None:
            with open_hdf5(process_file) as h5f:
                entries = BaseProcessInfo._get_process_nodes(
                    root_node=h5f, process=DataListener
                )
                if len(entries) == 0:
                    _logger.info(
                        "unable to find a DarkRef process in %s" % process_file
                    )
                    return None
                elif len(entries) > 1:
                    raise ValueError("several entry found, entry should be " "specify")
                else:
                    entry = list(entries.keys())[0]
                    _logger.info("take %s as default entry" % entry)

        with open_hdf5(process_file) as h5f:
            dl_nodes = BaseProcessInfo._get_process_nodes(
                root_node=h5f[entry], process=DataListener
            )
            index_to_path = {}
            for key, index in dl_nodes.items():
                index_to_path[index] = key

            if len(dl_nodes) == 0:
                return None
            # take the last processed dark ref
            last_process_index = sorted(dl_nodes.values())[-1]
            last_process_dl = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug(
                    "several processing found for data listener,"
                    "take the last one: %s" % last_process_dl
                )

            if "configuration" in h5f[last_process_dl].keys():
                results_node = h5f[last_process_dl]["configuration"]
                if "file_proposal" in results_node.keys():
                    try:
                        fp = h5py_read_dataset(
                            results_node["file_proposal"]
                        )  # .decode("UTF-8")
                    except AttributeError:
                        return None
                    else:
                        return fp
                else:
                    return None
            return None

    # TODO the 3 static functions get_proposal_file, get_sample_file and
    # get_source_scans are sharing a lot of source code and should
    # be 'concatenate'
    @staticmethod
    def get_sample_file(process_file, entry):
        """Return the proposal file of the experimentation if registred by the
        data listener"""
        if entry is None:
            with open_hdf5(process_file) as h5f:
                entries = BaseProcessInfo._get_process_nodes(
                    root_node=h5f, process=DataListener
                )
                if len(entries) == 0:
                    _logger.info(
                        "unable to find a DarkRef process in %s" % process_file
                    )
                    return None
                elif len(entries) > 1:
                    raise ValueError("several entry found, entry should be " "specify")
                else:
                    entry = list(entries.keys())[0]
                    _logger.info("take %s as default entry" % entry)

        with open_hdf5(process_file) as h5f:
            dl_nodes = BaseProcessInfo._get_process_nodes(
                root_node=h5f[entry], process=DataListener
            )
            index_to_path = {}
            for key, value in dl_nodes.items():
                index_to_path[key] = key

            if len(dl_nodes) == 0:
                return None
            # take the last processed dark ref
            last_process_index = sorted(dl_nodes.keys())[-1]
            last_process_dl = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug(
                    "several processing found for data listener,"
                    "take the last one: %s" % last_process_dl
                )

            if "configuration" in h5f[last_process_dl].keys():
                results_node = h5f[last_process_dl]["configuration"]
                if "sample_file" in results_node.keys():
                    return h5py_read_dataset(
                        results_node["sample_file"]
                    )  # .decode("UTF-8")
                else:
                    return None
            return None

    @staticmethod
    def get_master_sample_file(process_file, entry):
        """Return the proposal file of the experimentation if registered by the
        data listener"""
        if entry is None:
            with open_hdf5(process_file) as h5f:
                entries = BaseProcessInfo._get_process_nodes(
                    root_node=h5f, process=DataListener
                )
                if len(entries) == 0:
                    _logger.info(
                        "unable to find a DarkRef process in %s" % process_file
                    )
                    return None
                elif len(entries) > 1:
                    raise ValueError("several entry found, entry should be " "specify")
                else:
                    entry = list(entries.keys())[0]
                    _logger.info("take %s as default entry" % entry)

        with open_hdf5(process_file) as h5f:
            dl_nodes = BaseProcessInfo._get_process_nodes(
                root_node=h5f[entry], process=DataListener
            )
            index_to_path = {}
            for key, value in dl_nodes.items():
                index_to_path[key] = key

            if len(dl_nodes) == 0:
                return None
            # take the last processed dark ref
            last_process_index = sorted(dl_nodes.keys())[-1]
            last_process_dl = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug(
                    "several processing found for data listener,"
                    "take the last one: %s" % last_process_dl
                )

            if "configuration" in h5f[last_process_dl].keys():
                results_node = h5f[last_process_dl]["configuration"]
                if "master_sample_file" in results_node.keys():
                    return h5py_read_dataset(
                        results_node["master_sample_file"]
                    )  ##.decode("UTF-8")
                else:
                    return None
            return None

    @staticmethod
    def get_source_scans(process_file, entry):
        """Return the list of 'bliss scan directory' created for holding this
        specific sequence data
        """
        if entry is None:
            with open_hdf5(process_file) as h5f:
                entries = BaseProcessInfo._get_process_nodes(
                    root_node=h5f, process=DataListener
                )
                if len(entries) == 0:
                    _logger.info(
                        "unable to find a DarkRef process in %s" % process_file
                    )
                    return None
                elif len(entries) > 1:
                    raise ValueError("several entry found, entry should be " "specify")
                else:
                    entry = list(entries.keys())[0]
                    _logger.info("take %s as default entry" % entry)

        with open_hdf5(process_file) as h5f:
            dl_nodes = BaseProcessInfo._get_process_nodes(
                root_node=h5f[entry], process=DataListener
            )
            index_to_path = {}
            for key, index in dl_nodes.items():
                index_to_path[index] = key

            if len(dl_nodes) == 0:
                return {}
            # take the last processed dark ref
            last_process_index = sorted(dl_nodes.values())[-1]
            last_process_dl = index_to_path[last_process_index]
            if (len(index_to_path)) > 1:
                _logger.debug(
                    "several processing found for data listener,"
                    "take the last one: %s" % last_process_dl
                )

            if "configuration" in h5f[last_process_dl].keys():
                results_node = h5f[last_process_dl]["configuration"]
                if "source_scans" in results_node.keys():
                    tmp_res = h5py_read_dataset(results_node["source_scans"])

                    def conv(my_str):
                        if hasattr(my_str, "decode"):
                            return my_str.decode("UTF-8")
                        else:
                            return my_str

                    res = [conv(mystr) for mystr in tmp_res]
                    return res
                else:
                    return None
            return None

    def convert(self, bliss_file: str, entry: str) -> tuple:
        """

        :param bliss_file: file to be converted
        :param entry: entry in the tango .h5 file to be converter
        :return: tuple (output file, output file entry). Both are set to None
                 if conversion fails.
        """
        from tomwer.core.process.control.nxtomomill import (
            H5ToNxProcess as NxTomomillProcess,
        )

        conf = self.get_configuration()
        if conf is None:
            conf = {}
        configuration = HDF5Config.from_dict(conf)
        if configuration.output_file is None:
            output_file_path = NxTomomillProcess.deduce_output_file_path(
                bliss_file,
                entry=entry,
                outputdir=None,
                scan=BlissScan(
                    master_file=bliss_file, entry=entry, proposal_file=bliss_file
                ),
            )
        else:
            output_file_path = configuration.output_file

        if os.path.exists(output_file_path):
            if not self._ask_user_for_overwritting(output_file_path):
                return ((None, None),)

        if entry.startswith("/"):
            entries = (entry,)
        else:
            entries = ("/" + entry,)

        # work around: we need to wait a bit before converting the file
        # otherwise it looks like the file might not be ended to be
        # write
        def sequence_is_finished():
            try:
                with open_hdf5(bliss_file) as h5f:
                    end_scan_path = "/".join((entry, "end_time"))
                    return end_scan_path in h5f
            except Exception:
                return False

        timeout = self.TIMOUT_READ_FILE

        while timeout > 0 and not sequence_is_finished():
            timeout -= 0.2
            time.sleep(0.2)

        if timeout <= 0:
            _logger.error(f"unable to access {entry}@{bliss_file}. (Write never ended)")
            return ((None, None),)
        # one more delay to insure we can read it. Some frame might need more time to be dump from lima.
        time.sleep(4)

        # force some parameters
        configuration.input_file = bliss_file
        configuration.output_file = output_file_path
        configuration.entries = entries
        configuration.single_file = False
        configuration.overwrite = True
        configuration.request_input = True
        configuration.file_extension = ".nx"
        try:
            convs = nxtomomill_converter.from_h5_to_nx(
                configuration=configuration,
                input_callback=nxtomomill_input_callback,
                progress=None,
            )
        except Exception as e:
            _logger.error(
                f"Fail to convert from tango file: {bliss_file} to NXTomo. Error is: {e}"
            )
            return ((None, None),)
        else:
            return convs

    def _ask_user_for_overwritting(self, file_path):
        res = None
        while res not in ("Y", "n"):
            res = input(
                "The process will overwrite %s. Do you agree ? (Y/n)" "" % file_path
            )

        return res == "Y"

    def _signal_scan_ready(self, scan):
        assert isinstance(scan, NXtomoScan)
        pass

    def activate(self, activate=True):
        """
        activate or deactivate the thread. When deactivate call join and
        delete the thread

        :param activate:
        """
        if activate:
            if self._listening_thread is not None:
                _logger.info("listening is already activate")
            else:
                assert isinstance(self, DataListener)
                if not self.is_port_available():
                    raise OSError("Port already used")
                self.get_listening_thread().start()
        else:
            if self._listening_thread is None:
                return
            else:
                self.delete_listening_thread()

    def is_active(self):
        if self._listening_thread is None:
            return False
        else:
            return True

    def is_port_available(self):
        """

        :return: True if the port is available else False
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return not s.connect_ex((self._host, self._port)) == 0



from tomwer.core.process.task import Task
from tomwer.core.utils.scanutils import data_identifier_to_scan


class SingleTomoObjProcess(
    Task, optional_input_names=("tomo_obj",), output_names=("tomo_obj",)
):
    """For now data can only be a single element and not a list.
    This must be looked at.
    Also when part of an ewoks graph 'data' is mandatory which is not the class
    when part of a orange workflow. Those can be added interactively"""

    def run(self):
        self.outputs.tomo_obj = data_identifier_to_scan(self.inputs.tomo_obj)



from __future__ import annotations

import logging

from silx.gui import qt

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Output, OWBaseWidget

from tomwer.core.cluster import SlurmClusterConfiguration
from tomwer.gui.cluster.slurm import SlurmSettingsWindow

_logger = logging.getLogger(__name__)


class SlurmClusterOW(OWBaseWidget, openclass=True):
    """
    Orange widget to define a slurm cluster as input of other
    widgets (based on nabu for now)
    """

    name = "slurm cluster"
    id = "orange.widgets.tomwer.cluster.SlurmClusterOW.SlurmClusterOW"
    description = "Let the user configure the cluster to be used."
    icon = "icons/slurm.svg"
    priority = 20
    keywords = ["tomography", "tomwer", "slurm", "cluster", "configuration"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    _ewoks_default_inputs = Setting(dict())

    class Outputs:
        config_out = Output(name="cluster_config", type=SlurmClusterConfiguration)

    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        layout = gui.vBox(self.mainArea, self.name).layout()
        self._widget = SlurmSettingsWindow(parent=self)
        self._widget.setWindowFlags(qt.Qt.Widget)
        layout.addWidget(self._widget)

        if self._ewoks_default_inputs != {}:
            self._widget.setConfiguration(self._ewoks_default_inputs)

        # trigger the signal to avoid any user request
        self.Outputs.config_out.send(self.getConfiguration())

        # connect signal / slot
        self._widget.sigConfigChanged.connect(self._configurationChanged)

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("_static_input", None)
        if static_input not in (None, {}):
            _logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    def _configurationChanged(self):
        slurmClusterconfiguration = self.getConfiguration()
        self.Outputs.config_out.send(slurmClusterconfiguration)
        self._ewoks_default_inputs = slurmClusterconfiguration.to_dict()

    def getConfiguration(self):
        return self._widget.getSlurmClusterConfiguration()



from tomwer.core.utils.deprecation import deprecated_warning

deprecated_warning(
    "module",
    name="tomwer.app.darkref",
    reason="Has been moved",
    replacement="tomwer.app.reducedarkflat",
    only_once=True,
)
from tomwer.app.reducedarkflat import *  # noqa F401



# coding: utf-8

import pytest
from nxtomo.nxobject.nxdetector import ImageKey

from tomwer.core.utils.nxtomoutils import get_n_series


def test_get_n_series():
    """test tomwer.core.utils.nxtomoutils.get_n_series function"""
    array_1 = (
        [ImageKey.DARK_FIELD] * 2
        + [ImageKey.FLAT_FIELD]
        + [ImageKey.PROJECTION] * 4
        + [ImageKey.FLAT_FIELD]
    )

    with pytest.raises(ValueError):
        get_n_series(array_1, ImageKey.INVALID)
    with pytest.raises(ValueError):
        get_n_series(array_1, 3)

    assert len(array_1) == 8
    assert get_n_series(array_1, ImageKey.DARK_FIELD) == 1
    assert get_n_series(array_1, ImageKey.FLAT_FIELD) == 2
    assert get_n_series(array_1, ImageKey.PROJECTION) == 1
    assert get_n_series(array_1, 0) == 1

    array_2 = (
        [ImageKey.FLAT_FIELD.value]
        + [ImageKey.PROJECTION.value] * 4
        + [ImageKey.INVALID.value] * 2
        + [ImageKey.PROJECTION.value] * 3
        + [ImageKey.FLAT_FIELD.value]
    )
    assert get_n_series(array_2, ImageKey.DARK_FIELD) == 0
    assert get_n_series(array_2, ImageKey.PROJECTION) == 1
    assert get_n_series(array_2, ImageKey.FLAT_FIELD) == 2



"""Define Interface for widget handling stitching over one axis"""

from __future__ import annotations

from silx.gui import qt


class _SingleAxisMixIn:
    """Common API for widget able to handle stitching over one axis (no matter if it is 0, 1, 2 (aka z, y, x))"""

    @property
    def first_axis(self) -> int:
        """
        Axis along which stitching is done. For stitching along z will return 0...
        """
        return self._axis

    @property
    def second_axis(self) -> int:
        """
        second axis involved in the stitching. In case of stitching along axis 0 the axis 1can also be involved (axis present in the camera reference)
        """
        if self._axis == 0:
            return 1
        elif self._axis == 1:
            return 0
        else:
            raise NotImplementedError

    @property
    def third_axis(self):
        all_axis = {(0, 1, 2)}
        all_axis.remove(self.first_axis)
        all_axis.remove(self.second_axis)
        return all_axis.pop()

    @staticmethod
    def axis_alias(axis: int):
        if axis == 0:
            return "z"
        elif axis == 1:
            return "y"
        elif axis == 2:
            return "z"
        else:
            raise ValueError(f"axis should be in (0, 1, 2). {axis} given")


class SingleAxisMetaClass(type(qt.QMainWindow), type(_SingleAxisMixIn)):
    """
    Metaclass for single axis stitcher in order to aggregate dumper class and axis
    """

    def __new__(mcls, name, bases, attrs, axis: int | None = None):
        mcls = super().__new__(mcls, name, bases, attrs)
        mcls._axis = axis  # used from _SingleAxisMixIn interface
        return mcls



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Application to edit (key by key) the 'imagekey' value of a NXtomo ('imagekey' defines if a frame is a 'dark', 'flat' or a 'projection')
"""
import argparse
import logging
import signal
import sys

from silx.gui import qt

from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.utils.resource import increase_max_number_file
from tomwer.core.process.edit.imagekeyeditor import ImageKeyEditorTask
from tomwer.gui import icons
from tomwer.gui.edit.imagekeyeditor import ImageKeyDialog as _ImageKeyDialog
from tomwer.gui.utils.splashscreen import getMainSplashScreen

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


class ImageKeyDialog(_ImageKeyDialog):
    def __init__(self, *args, **kwargs):
        _ImageKeyDialog.__init__(self, *args, **kwargs)
        self._scan = None

    def setScan(self, scan):
        if not isinstance(scan, NXtomoScan):
            raise TypeError("This only manage NXtomoScan")
        self._scan = scan
        _ImageKeyDialog.setScan(self, scan)

    def _validateReleased(self):
        if not self._scan:
            raise ValueError("Scan should be set first")

        task = ImageKeyEditorTask(
            inputs={
                "data": self._scan,
                "configuration": {"modifications": self.getModifications()},
            }
        )
        try:
            task.run()
        except Exception as e:
            _logger.error(e)
        else:
            print("edition succeed")


def getinputinfo():
    return "tomwer image-key-editor [scan_path] [entry]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


def main(argv):
    import os

    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("scan_path", help="HDF5File to edit")
    parser.add_argument("entry", default=None, help="Entry to treat")
    options = parser.parse_args(argv[1:])

    # image key can only handle NXtomoScan for now
    scan = NXtomoScan(scan=options.scan_path, entry=options.entry)
    increase_max_number_file()

    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler
    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    splash = getMainSplashScreen()
    options.scan_path = options.scan_path.rstrip(os.path.sep)

    widget = ImageKeyDialog(parent=None)
    widget.setWindowFlags(qt.Qt.Window)
    widget.setScan(scan)
    widget.setWindowTitle("Image key editor")
    widget.setWindowIcon(icons.getQIcon("tomwer"))
    splash.finish(widget)
    widget.show()
    app.exec_()


if __name__ == "__main__":
    main(sys.argv)



import gc
import os
import shutil
import tempfile

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.control.DataValidatorOW import DataValidatorOW
from tomwer.core.utils.scanutils import MockNXtomo


class TestDataValidatorOW(TestCaseQt):
    """
    simple test on the DataValidatorOW widget
    """

    def setUp(self):
        super().setUp()
        self.window = DataValidatorOW()
        self.folder = tempfile.mkdtemp()
        self.scan_1 = MockNXtomo(
            scan_path=os.path.join(self.folder, "scan_1"),
            n_proj=20,
            n_ini_proj=20,
            dim=10,
        ).scan
        self.scan_2 = MockNXtomo(
            scan_path=os.path.join(self.folder, "scan_2"),
            n_proj=20,
            n_ini_proj=20,
            dim=10,
        ).scan

    def tearDown(self):
        self.window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.window.close()
        self.window = None
        shutil.rmtree(self.folder)
        gc.collect()

    def test(self):
        def wait_processing():
            while self.qapp.hasPendingEvents():
                self.qapp.processEvents()

        assert self.window.getNScanToValidate() == 0
        self.window.addScan(self.scan_1)
        wait_processing()
        self.window.addScan(self.scan_2)
        wait_processing()
        assert self.window.getNScanToValidate() == 2
        self.window._widget.validateCurrentScan()
        wait_processing()
        assert self.window.getNScanToValidate() == 1



# coding: utf-8


from tomwer.core.process.visualization.sinogramviewer import _SinogramViewerPlaceHolder


def test_sinogram_viewer():
    process = _SinogramViewerPlaceHolder(
        inputs={
            "data": None,
        }
    )
    process.run()



from silx.gui import qt

from tomwer.core.process.stitching.metadataholder import (
    StitchingMetadata as _StitchingMetadata,
)


class QStitchingMetadata(qt.QObject, _StitchingMetadata):
    """
    overload of a TomoObject to register metadata set by the user on positions
    """

    sigChanged = qt.Signal()
    """
    emit when some parameter changed
    """

    def __init__(self, parent=None, tomo_obj=None) -> None:
        super().__init__(parent)
        _StitchingMetadata.__init__(self=self, tomo_obj=tomo_obj)

    def setPixelOrVoxelSize(self, value, axis):
        super().setPixelOrVoxelSize(value=value, axis=axis)
        self.sigChanged.emit()

    def setPxPos(self, value, axis):
        super().setPxPos(value=value, axis=axis)
        self.sigChanged.emit()

    def setMetricPos(self, value, axis):
        super().setMetricPos(value=value, axis=axis)
        self.sigChanged.emit()



# coding: utf-8
from __future__ import annotations


from ewokscore.task import Task as EwoksTask


class _DataViewerPlaceHolder(EwoksTask, input_names=("data",)):
    """
    Task to browse a scan and associated processing (display raw frames, normalized projections, recosntructed slices...)
    """

    def run(self):
        pass



# coding: utf-8
from __future__ import annotations


from silx.gui import qt


class _IgnoreWheelBase:
    """QWidget ignoring "wheelEvent" and forwarding it to its parent in case this is a QScrollArea"""

    def wheelEvent(self, *args, **kwargs):
        pass


class QComboBoxIgnoreWheel(qt.QComboBox, _IgnoreWheelBase):
    pass


class QDoubleSpinBoxIgnoreWheel(qt.QDoubleSpinBox):
    pass


class QSpinBoxIgnoreWheel(qt.QSpinBox, _IgnoreWheelBase):
    pass



# coding: utf-8
from __future__ import annotations


import gc
import os
import pickle
import shutil
import tempfile
from time import sleep

from nabu.pipeline.config import get_default_nabu_config
from nabu.pipeline.fullfield.nabu_config import (
    nabu_config as nabu_fullfield_default_config,
)
from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.reconstruction.NabuVolumeOW import NabuVolumeOW
from tomwer.core.utils.scanutils import MockNXtomo


class TestNabuVolumeOW(TestCaseQt):
    def setUp(self):
        super().setUp()
        self.tmp_dir = tempfile.mkdtemp()
        self.scan = MockNXtomo(
            scan_path=os.path.join(self.tmp_dir, "myscan"),
            n_proj=20,
            n_ini_proj=20,
            dim=10,
        ).scan
        self.widget = NabuVolumeOW()
        self.widget._processingStack.setDryRun(True)

    def tearDown(self):
        self.widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.widget.close()
        self.widget = None
        shutil.rmtree(self.tmp_dir)
        self.qapp.processEvents()
        gc.collect()

    def test_serializing(self):
        pickle.dumps(self.widget.getConfiguration())

    def test_literal_dumps(self):
        literal_dumps(self.widget.getConfiguration())

    def test_scan_as_None(self):
        "test if scan is None"
        self.widget.process(None)

    def test_scan_un_preprocessed(self):
        "test when scan is unconfigured (No nabu_recons_params defined)"
        assert self.scan.nabu_recons_params is None
        self.widget.process(self.scan)

    def test_scan_ready_to_be_processed(self):
        "test if scan has valid reconstruction parameters to be runned"
        self.scan.nabu_recons_params = get_default_nabu_config(
            nabu_fullfield_default_config
        )
        self.widget.process(self.scan)
        timeout = 3
        loop_sleep_time = 0.05
        while not self.widget._processingStack.can_process_next():
            self.qapp.processEvents()
            timeout -= loop_sleep_time
            if timeout <= 0.0:
                raise TimeoutError("volume not process within expected time")
            else:
                sleep(loop_sleep_time)



"""contains 'QConfigFileDialog': dialog to get / set a .cfg file
"""

import os
from silx.gui import qt


class QConfigFileDialog(qt.QFileDialog):
    """dialog to get / set a configuration file (as a .cfg file)"""

    def __init__(self, parent, multiSelection=False):
        qt.QFileDialog.__init__(self, parent)
        self.setNameFilters(
            [
                "config file (*.cfg *.conf *.config *.par)",
                "Any files (*)",
            ]
        )
        self._selected_files = []

        # check if 'TOMWER_DEFAULT_INPUT_DIR' has been set
        if os.environ.get("TOMWER_DEFAULT_INPUT_DIR", None) and os.path.exists(
            os.environ["TOMWER_DEFAULT_INPUT_DIR"]
        ):
            self.setDirectory(os.environ["TOMWER_DEFAULT_INPUT_DIR"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""Application to append dark of flat frames to an existing NXtomo
"""

import argparse
import logging
import signal
import sys

from nxtomomill.utils import add_dark_flat_nx_file
from silx.gui import qt

from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.utils.resource import increase_max_number_file
from tomwer.gui import icons
from tomwer.gui.edit.dkrfpatch import DarkRefPatchWidget
from tomwer.gui.utils.splashscreen import getMainSplashScreen

logging.basicConfig(level=logging.WARNING)
_logger = logging.getLogger(__name__)


class _DarkRefPatchDialog(qt.QDialog):
    def __init__(self, parent):
        qt.QDialog.__init__(self, parent)
        self._scan = None
        self.setLayout(qt.QVBoxLayout())
        self._mainWidget = DarkRefPatchWidget(self)
        self.layout().addWidget(self._mainWidget)

        # buttons
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self._buttons.button(qt.QDialogButtonBox.Ok).setText("Apply")
        self.layout().addWidget(self._buttons)

        # connect signal slot
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self._apply)

    def _apply(self):
        if self._scan is None:
            raise ValueError("Scan should be set before applying patch")
        url_sd = self._mainWidget.getStartDarkUrl()
        url_sf = self._mainWidget.getStartFlatUrl()
        url_ed = self._mainWidget.getEndDarkUrl()
        url_ef = self._mainWidget.getEndFlatUrl()
        add_dark_flat_nx_file(
            file_path=self._scan.master_file,
            entry=self._scan.entry,
            darks_start=url_sd,
            flats_start=url_sf,
            darks_end=url_ed,
            flats_end=url_ef,
        )

    def setScan(self, scan):
        self._scan = scan


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path",
        help="HDF5 - Nexus file path",
        default=None,
    )
    parser.add_argument("entry", help="NXTomo entry to load", default=None)
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )

    options = parser.parse_args(argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    increase_max_number_file()

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])
    splash = getMainSplashScreen()
    qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    scan = ScanFactory.create_scan_object(
        scan_path=options.scan_path, entry=options.entry
    )

    # define the process_index is any tomwer_processes_existing
    scan.set_process_index_frm_tomwer_process_file()

    if options.debug:
        _logger.setLevel(logging.DEBUG)

    window = _DarkRefPatchDialog(parent=None)
    window.setScan(scan=scan)
    window.setWindowTitle("patch dark & flat")
    window.setWindowIcon(icons.getQIcon("tomwer"))

    splash.finish(window)
    window.show()
    qt.QApplication.restoreOverrideCursor()
    app.exec_()


def getinputinfo():
    return "tomwer darkref-patch [scanDir]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


if __name__ == "__main__":
    main(sys.argv)



from __future__ import annotations

import logging
import os
from copy import copy

from nxtomomill.io.config import TomoHDF5Config as HDF5Config
from orangewidget import gui
from silx.gui import qt
from silx.gui.utils import blockSignals

from orangecontrib.tomwer.orange.managedprocess import TomwerWithStackStack
from orangecontrib.tomwer.widgets.control.NXTomomillMixIn import NXTomomillMixIn
from tomwer.core.process.control.nxtomomill import H5ToNxProcess
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.nxtomoscan import NXtomoScan, NXtomoScanIdentifier
from tomwer.gui.control.datalist import BlissHDF5DataListMainWindow
from tomwer.gui.control.nxtomomill import NXTomomillInput, OverwriteMessage
from tomwer.core.process.output import ProcessDataOutputDirMode
from ewoksorange.bindings.owwidgets import invalid_data


logger = logging.getLogger(__name__)


class NXTomomillOW(
    TomwerWithStackStack,
    NXTomomillMixIn,
    ewokstaskclass=H5ToNxProcess,
):
    """
    Widget to allow user to pick some bliss files and that will convert them
    to HDF5scan.
    """

    name = "nxtomomill h52nx (bliss-HDF5)"
    id = "orange.widgets.tomwer.control.NXTomomillOW.NXTomomillOW"
    description = (
        "Read a bliss .h5 file and extract from it all possible"
        "NxTomo. When validated create a TomwerBaseScan for each "
        "file and entry"
    )
    icon = "icons/nxtomomill.svg"
    priority = 120
    keywords = [
        "hdf5",
        "nexus",
        "tomwer",
        "file",
        "convert",
        "NXTomo",
        "tomography",
        "nxtomomill",
        "h52nx",
    ]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    CONFIG_CLS = HDF5Config

    LOGGER = logger

    _ewoks_inputs_to_hide_from_orange = (
        "h5_to_nx_configuration",
        "progress",
        "serialize_output_data",
    )

    def __init__(self, parent=None):
        TomwerWithStackStack.__init__(self, parent=parent)
        NXTomomillMixIn.__init__(self)
        _layout = gui.vBox(self.mainArea, self.name).layout()

        self.widget = BlissHDF5DataListMainWindow(parent=self)
        _layout.addWidget(self.widget)
        self.__request_input = True
        # do we ask the user for input if missing
        self._inputGUI = None
        """Gui with cache for missing field in files to be converted"""
        self._canOverwriteOutputs = False
        """Cache to know if we have to ask user permission for overwriting"""

        # expose API
        self.n_scan = self.widget.n_scan
        # alias used for the 'simple workflow' for now
        self.start = self._sendAll

        # connect signal / slot
        self.widget._sendSelectedButton.clicked.connect(self._sendSelected)
        self.widget.sigNXTomoCFGFileChanged.connect(self._saveNXTomoCfgFile)
        self.widget.sigUpdated.connect(self._updateSettings)

        # set default configuration is no existing configuration file defined in the settings
        self.update_default_inputs(
            h5_to_nx_configuration=HDF5Config().to_dict(),
        )
        if isinstance(self.task_output_changed_callbacks, set):
            self.task_output_changed_callbacks.add(self._notify_state)
        elif isinstance(self.task_output_changed_callbacks, list):
            self.task_output_changed_callbacks.append(self._notify_state)
        else:
            raise NotImplementedError

        # handle settings
        self._loadSettings()

    def _updateSettings(self):
        self._scans = []
        for scan in self.widget.datalist._myitems:
            # kept for backward compatibility since 0.11. To be removed on the future version.
            if "@" in scan:
                entry, file_path = scan.split("@")
                nxtomo_scan = NXtomoScan(entry=entry, scan=file_path)
                self.add(nxtomo_scan)
            else:
                self._scans.append(scan)
        output_dir = self.widget.getOutputFolder()
        if isinstance(output_dir, ProcessDataOutputDirMode):
            output_dir = output_dir.value
        self._ewoks_default_inputs["output_dir"] = output_dir  # pylint: disable=E1137

    @property
    def request_input(self):
        return self.__request_input

    @request_input.setter
    def request_input(self, request):
        self.__request_input = request

    def get_task_inputs(self):
        return {
            "h5_to_nx_configuration": self.__configuration_cache.to_dict(),
            "serialize_output_data": False,
        }

    def handleNewSignals(self) -> None:
        """Invoked by the workflow signal propagation manager after all
        signals handlers have been called.
        """
        # for now we want to avoid propagation any processing.
        # task will be executed only when the user validates the dialog
        bliss_scan = super().get_task_inputs().get("bliss_scan", None)
        if bliss_scan is not None:
            if not isinstance(bliss_scan, BlissScan):
                raise TypeError("bliss_scan is expected to be an instance of BlissScan")
            self.add(bliss_scan.master_file)

    def _convertAndSend(self, bliss_url: str):
        """

        :param bliss_url: string at entry@file format
        """
        logger.processStarted(f"Start translate {bliss_url} to NXTomo")
        self.__configuration_cache = HDF5Config.from_dict(
            copy(self.get_default_input_values()["h5_to_nx_configuration"])
        )

        identifier = NXtomoScanIdentifier.from_str(bliss_url)
        bliss_scan = BlissScan(
            master_file=identifier.file_path,
            entry=identifier.data_path,
            proposal_file=None,
        )

        output_file_path = H5ToNxProcess.deduce_output_file_path(
            bliss_scan.master_file,
            entry=bliss_scan.entry,
            outputdir=self.widget.getOutputFolder(),
            scan=bliss_scan,
        )

        self.__configuration_cache.input_file = bliss_scan.master_file
        self.__configuration_cache.output_file = output_file_path
        self.__configuration_cache.entries = (bliss_scan.entry,)
        self.__configuration_cache.single_file = False
        self.__configuration_cache.overwrite = True
        self.__configuration_cache.request_input = self.request_input
        self.__configuration_cache.file_extension = ".nx"

        self._processBlissScan(bliss_scan)

    def _userAgreeForOverwrite(self, file_path):
        if self._canOverwriteOutputs:
            return True
        else:
            msg = OverwriteMessage(self)
            text = "NXtomomill will overwrite \n %s. Do you agree ?" % file_path
            msg.setText(text)
            if msg.exec_():
                self._canOverwriteOutputs = msg.canOverwriteAll()
                return True
            else:
                return False

    def _processBlissScan(self, bliss_scan):
        if bliss_scan is None:
            return
        output_file_path = H5ToNxProcess.deduce_output_file_path(
            bliss_scan.master_file,
            entry=bliss_scan.entry,
            outputdir=self.widget.getOutputFolder(),
            scan=bliss_scan,
        )
        # check user has rights to write on the folder
        dirname = os.path.dirname(output_file_path)
        if os.path.exists(dirname) and not os.access(dirname, os.W_OK):
            msg = qt.QMessageBox(self)
            msg.setIcon(qt.QMessageBox.Warning)
            text = f"You don't have write rights on '{dirname}'. Unable to generate the nexus file associated to {str(bliss_scan)}"
            msg.setWindowTitle("No rights to write")
            msg.setText(text)
            msg.show()
            return
        # check if need to overwrite the file
        elif os.path.exists(output_file_path):
            if not self._userAgreeForOverwrite(output_file_path):
                return

        # keep 'h5_to_nx_configuration' up to date according to input folder and output_file updates
        self.update_default_inputs(
            h5_to_nx_configuration=self.__configuration_cache.to_dict()
        )
        try:
            self._execute_ewoks_task(  # pylint: disable=E1123
                propagate=True,
                log_missing_inputs=False,
            )
        except Exception:
            self._execute_ewoks_task(propagate=True)  # pylint: disable=E1123, E1120

    def _loadSettings(self):
        with blockSignals(self.widget):
            for scan in self._scans:
                assert isinstance(scan, str)
                try:
                    self.widget.add(scan)
                except Exception as e:
                    logger.error(f"Fail to add {scan}. Error is {e}")
                else:
                    logger.warning(f"{scan} is an invalid link to a file")
            if (
                "nxtomomill_cfg_file" in self._ewoks_default_inputs
            ):  # pylint: disable=E1135
                nxtomo_cfg_file = self._ewoks_default_inputs[  # pylint: disable=E1136
                    "nxtomomill_cfg_file"
                ]
                self.widget.setCFGFilePath(nxtomo_cfg_file)
            if "output_dir" in self._ewoks_default_inputs:  # pylint: disable=E1135
                self.widget.setOutputFolder(
                    self._ewoks_default_inputs["output_dir"]  # pylint: disable=E1136
                )

    def getHDF5Config(self):
        configuration_file = self.widget.getCFGFilePath()
        if configuration_file in (None, ""):
            configuration = HDF5Config()
        else:
            try:
                configuration = HDF5Config.from_cfg_file(configuration_file)
            except Exception as e:
                logger.error(
                    f"Fail to use configuration file {configuration_file}. Error is {e}. No conversion will be done."
                )
                return None
        return configuration

    def _saveNXTomoCfgFile(self, cfg_file):
        super()._saveNXTomoCfgFile(cfg_file, keyword="h5_to_nx_configuration")

    def _sendSelected(self):
        """Send a signal for selected scans found to the next widget"""
        self._inputGUI = NXTomomillInput()
        # reset the GUI for input (reset all the cache for answers)
        self._canOverwriteOutputs = False
        for bliss_url in self.widget.datalist.selectedItems():
            data = bliss_url.data(qt.Qt.UserRole)
            assert isinstance(data, NXtomoScan)
            identifier = data.get_identifier()
            self._inputGUI.setBlissScan(
                entry=identifier.data_path, file_path=identifier.file_path
            )
            self._convertAndSend(identifier.to_str())

    def _sendAll(self):
        """Send a signal for each scan found to the next widget"""
        self._inputGUI = NXTomomillInput()
        # reset the GUI for input (reset all the cache for answers)
        self._canOverwriteOutputs = False
        for bliss_url in self.widget.datalist._myitems.values():
            data = bliss_url.data(qt.Qt.UserRole)
            identifier = data.get_identifier()
            assert isinstance(data, NXtomoScan)
            self._inputGUI.setBlissScan(
                entry=identifier.data_path, file_path=identifier.file_path
            )
            self._convertAndSend(identifier.to_str())

    def _notify_state(self):
        try:
            task_executor = self.sender()
            task_suceeded = task_executor.succeeded
            config = task_executor.current_task.inputs.h5_to_nx_configuration
            config = HDF5Config.from_dict(config)
            scan = task_executor.current_task.outputs.data
            if task_suceeded:
                self.notify_succeed(scan=scan)
            else:
                self.notify_failed(scan=scan)
        except Exception as e:
            logger.error(f"failed to handle task finished callback. Reason is {e}")

    def trigger_downstream(self) -> None:
        for ewoksname, var in self.get_task_outputs().items():
            # note: for now we want to trigger 'data' for each items of 'datas'
            if ewoksname == "series" and not (
                invalid_data.is_invalid_data(var.value) or var.value is None
            ):
                for data in var.value:
                    data_channel = self._get_output_signal("data")
                    data_channel.send(data)
                serie_channel = self._get_output_signal("series")
                # then send the list of value / series (also know as datas)
                serie_channel.send(var.value)
            elif ewoksname == "data" and not (
                invalid_data.is_invalid_data(var.value) or var.value is None
            ):
                pass  # handle by 'series' in this case

    def keyPressEvent(self, event):
        # forward Ctrl+A to the list as the shift ease selection of all
        modifiers = event.modifiers()
        key = event.key()
        if key == qt.Qt.Key_A and modifiers == qt.Qt.KeyboardModifier.ControlModifier:
            self.widget._widget.datalist.keyPressEvent(event)
        super().keyPressEvent(event)



# coding: utf-8

"""
This module analyze headDir data directory to detect scan to be reconstructed
"""
from __future__ import annotations

import logging
import os
import subprocess
from glob import glob

from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.signal import Signal

from .status import OBSERVATION_STATUS

logger = logging.getLogger(__name__)

try:
    from tomwer.synctools.rsyncmanager import RSyncManager

    has_rsync = False
except ImportError:
    logger.warning("rsyncmanager not available")
    has_rsync = True


"""
Useful tools
"""


class _DataWatcherProcess:
    """
    DataWatcherProcess is the process managing the observation of acquisition.
    Since we want to loop infinitly on a root folder we have to ignore the
    folder we previously detected. Otherwise if those folders are not removed
     we will loop infinitly. That is why we now have the ignoredFolders
     parameter.

    Example of usage of srcPattern and destPattern:

        For example during acquisition in md05 acquisition files are stored
        in /lbsram/data/visitor/x but some information (as .info) files are
        stored in /data/visitor/x.
        So we would like to check information in both directories.
        Furthermore we would like that all file not in /data/visitor/x will be
        copied as soon as possible into /data/visitor/x (using RSyncManager)

        To do so we can define a srcPattern ('/lbsram' in our example) and
        destPattern : a string replacing to srcPattern in order to get both
        repositories. ('' in out example)
        If srcPattern or destPattern are setted to None then we won't apply
        this 'two directories' synchronization and check

    :param  dataDir: Root directory containing data
    :param  waitXML: if True then we will be waiting for the XML of
        the acquisition to be writted. Otherwise we will look for the .info
        file and wait until all file will be copied
    :param srcPattern: the pattern to change by destPattern.
    :param destPattern: the pattern that will replace srcPattern in the
        scan path
    :param list ignoredFolders: the list of folders to ignored on parsing
                                 (them and sub folders)
    """

    sigNbDirExplored = Signal(int)
    """Signal emitted to notice the number of directory at the top level"""
    sigAdvanceExploration = Signal(int)
    """signal emitted each time a top level directory is parsed"""
    sigNewObservation = Signal(tuple)
    """used to signal a new step on the acquisition (first element of the tuple
        should be on eof OBSERVATION_STATUS, second element can be extra
        information to be displayed )"""

    sigNewInformation = Signal(str)
    """Signal used to communicate some random information (will originally be
        displayed in DataWatcherOW)"""

    INITIAL_STATUS = "not processing"

    def __init__(
        self,
        dataDir: str,
        srcPattern: str | None = None,
        destPattern: str | None = None,
    ):
        super().__init__()
        self.RootDir = os.path.abspath(dataDir)
        self.parsing_dir = os.path.abspath(dataDir)
        self.oldest = 0
        self.curdir = ""
        self.scan_name = ""
        self.scan_completed = False
        self.reconstructed = False
        self.status = self.INITIAL_STATUS
        """contains the step of acquisition we are looking for and the
        status on this step"""
        self.quitting = False
        self._removed = None

        srcPatternInvalid = srcPattern not in [None, ""] and not os.path.isdir(
            srcPattern
        )
        destPatternInvalid = destPattern not in [None, ""] and not os.path.isdir(
            destPattern
        )

        if srcPatternInvalid or destPatternInvalid:
            srcPattern = None
            destPattern = None

        self.srcPattern = srcPattern if destPattern is not None else None
        self.destPattern = destPattern if srcPattern is not None else None

    def _setStatus(self, status, info=None):
        assert status in OBSERVATION_STATUS
        self.status = status
        if info is None:
            self.sigNewObservation.emit((status,))
        else:
            self.sigNewObservation.emit((status, info))

    def _removeAcquisition(self, scanID, reason):
        raise NotImplementedError("Base class")

    def dir_explore(self):
        """
        Explore directory tree until valid .file_info_ext file is found.
        Tree explored by ascending order relative to directory date depending
        self.oldest

        :return: True if the acquisition as started else False
        """
        self.status = "start acquisition"

    def _sync(self):
        """Start to copy files from /lbsram/path to path if on lbsram

        :return: True if the synchronization starts
        """
        if has_rsync:
            return False

        if self.srcPattern:
            assert os.path.isdir(self.srcPattern) or self.srcPattern == ""
            assert os.path.isdir(self.destPattern) or self.destPattern == ""
            if self.RootDir.startswith(self.srcPattern):
                source = os.path.join(self.RootDir, self.parsing_dir)
                target = source.replace(self.srcPattern, self.destPattern, 1)
                info = f"Start synchronization between {source} and {target}"
                self.sigNewInformation.emit(info)

                target_dirname = os.path.dirname(target)
                # check if a synchronization is already running, avoid
                # duplication
                try:
                    if not RSyncManager().has_active_sync(source, target_dirname):
                        RSyncManager().sync_file(source, target_dirname)
                except subprocess.TimeoutExpired:
                    # if time out expired, do the synchronization anyway
                    RSyncManager().sync_file(source, target_dirname)
                return True
        else:
            return False

    def is_data_complete(self):
        """
        Check that data file is found and complete

        :return: - 0 if the acquisition is no finished
                 - 1 if the acquisition is finished
                 - -1 observation has been stopped
        """
        raise NotImplementedError("_DataWatcherProcess is a pure virtual class")

    def getCurrentDir(self):
        """Return the current dircetory parsed absolute path"""
        return os.path.join(self.RootDir, self.parsing_dir)

    def _isScanDirectory(self, directory):
        if EDFTomoScan.directory_contains_scan(
            directory, src_pattern=self.srcPattern, dest_pattern=self.destPattern
        ):
            scan = EDFTomoScan(directory)

            # keep compatibility EDF: get detector
            aux = directory.split(os.path.sep)
            infoname = os.path.join(directory, aux[-1] + EDFTomoScan.INFO_EXT)

            if self.srcPattern:
                infoname = infoname.replace(self.srcPattern, self.destPattern, 1)
            """
            assume 1st it is standard info filename
            """
            gd = glob(os.path.join(directory, infoname))
            if len(gd) == 0:
                """
                To detect if it is PCO like, check if info filename contains 0000
                """
                gd = glob(os.path.join(directory, EDFTomoScan.INFO_EXT))
                if len(gd) > 0:
                    self.detector = "Dimax"
        else:
            return False

        if scan.is_abort(src_pattern=self.srcPattern, dest_pattern=self.destPattern):
            self._removeAcquisition(
                scanID=directory, reason="acquisition aborted by the user"
            )

        # TODO: add a scan.get_status() function to get the acquisition status?
        self._setStatus("started")
        return True

    def is_abort(self):
        raise NotImplementedError("Base class")


class _DataWatchEmpty(_DataWatcherProcess):
    """A data watcher which just look under sub directories"""

    def is_data_complete(self):
        return False

    def _removeAcquisition(self, scanID, reason):
        pass



import os
import shutil
import tempfile

from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from tomwer.core.process.stitching.metadataholder import StitchingMetadata
from tomwer.gui.stitching.metadataholder import QStitchingMetadata
from tomwer.gui.stitching.tests.utils import create_scans_z_series
from tomwer.gui.stitching.config.tomoobjdetails import TomoObjectPositionInfos
from tomwer.gui.stitching.axisorderedlist import AxisOrderedTomoObjsModel


class TestTomoObjectPositionInfos(TestCaseQt):
    """
    Test edition of the tomo object position information
    """

    def setUp(self):
        super().setUp()
        self._tmp_path = tempfile.mkdtemp()
        self._y_pixel_size = 0.002
        self._x_pixel_size = 0.001
        self.scans = create_scans_z_series(
            self._tmp_path,
            z_positions_m=(0.200, 0.205, 0.210),
            x_positions_m=(0.0, 0.0, None),
            pixel_size=(self._y_pixel_size, self._x_pixel_size),
            raw_frame_width=100,
        )
        self.scan_0_metadata = QStitchingMetadata(tomo_obj=self.scans[0])
        assert self.scan_0_metadata.get_raw_position_m(axis=0) == 0.20
        assert self.scan_0_metadata.get_raw_position_m(axis=2) == 0.0
        self.scan_2_metadata = QStitchingMetadata(tomo_obj=self.scans[2])
        assert self.scan_2_metadata.get_raw_position_m(axis=0) == 0.21
        self._widget = TomoObjectPositionInfos()

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        shutil.rmtree(self._tmp_path)

    def testEdition(self):
        # test QStitchingMetadata are correctly read
        self._widget.show()
        self._widget.setStitchingMetadata(self.scan_0_metadata)
        self.qapp.processEvents()
        assert self._widget._axis_0_pos.get_final_px_position() == int(0.20 / 0.002)
        assert self._widget._axis_2_pos.get_final_px_position() == 0
        self._widget.setStitchingMetadata(self.scan_2_metadata)
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        assert self._widget._axis_0_pos.get_final_px_position() == int(0.21 / 0.002)
        assert self._widget._axis_2_pos.get_final_px_position() is None
        # test editing some parameters from the GUI. Make sure GUI is updated as the underlying metadata object
        self._widget._axis_0_pos._overrideMetricPositionCB.setChecked(True)
        self._widget._axis_0_pos.setRawMetricPosition(
            position_m=0.4, displayed_unit="nm"
        )
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        assert self._widget._axis_0_pos.get_final_px_position() == int(
            0.4 * 10e-6 / 0.002
        )
        self._widget._axis_0_pos._overridePixelPositionCB.setChecked(True)
        self._widget._axis_0_pos.setOverrridePxPosition(position_px=120)
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        assert self._widget._axis_0_pos.get_final_px_position() == int(120)
        assert self.scan_2_metadata.get_abs_position_px(axis=0) == int(120)


class TestAxisOrderedTomoObjsModel(TestCaseQt):
    """
    test the behavior of the z ordered model used for stitching
    """

    def setUp(self):
        super().setUp()
        self._tmp_path = tempfile.mkdtemp()
        self.scans_with_pos_and_pixel_size = create_scans_z_series(
            os.path.join(self._tmp_path, "case1"),
            z_positions_m=(0.200, 0.205, 0.210),
            pixel_size=0.001,
            raw_frame_width=100,
        )
        self.scans_with_pixel_size = create_scans_z_series(
            os.path.join(self._tmp_path, "case2"),
            z_positions_m=(None, None, None),
            pixel_size=0.001,
            raw_frame_width=100,
        )
        self.scans_without_metadata = create_scans_z_series(
            os.path.join(self._tmp_path, "case3"),
            z_positions_m=(None, None, None),
            pixel_size=None,
            raw_frame_width=100,
        )
        self._widget = qt.QTableView()
        self._widget.setModel(AxisOrderedTomoObjsModel(parent=self._widget, axis=0))

    def tearDown(self):
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._widget.close()
        self._widget = None
        while self.qapp.hasPendingEvents():
            self.qapp.processEvents()
        shutil.rmtree(self._tmp_path)

    def testWithMetadata(self):
        """
        make sure z ordering works in the best case scenario
        """
        first_expected_scan = (
            self.scans_with_pos_and_pixel_size[2].get_identifier().short_description()
        )
        # try giving it ordered
        for scan in self.scans_with_pos_and_pixel_size[::-1]:
            self._widget.model().addTomoObj(scan)
        assert (
            self._widget.model().data(
                self._widget.model().index(0, 1), qt.Qt.DisplayRole
            )
            == first_expected_scan
        )
        self._widget.model().clearTomoObjs()
        # try giving it ordered (but inverted)
        for scan in self.scans_with_pos_and_pixel_size:
            self._widget.model().addTomoObj(scan)
        assert (
            self._widget.model().data(
                self._widget.model().index(0, 1), qt.Qt.DisplayRole
            )
            == first_expected_scan
        )

    def testWithPixelSizeOnly(self):
        """
        make sure z ordering works when we are only aware about pixel size
        """
        for scan in self.scans_with_pixel_size[::-1]:
            self._widget.model().addTomoObj(scan)

        # try adding some metadata to object
        scans = self._widget.model()._objs
        scans = self._widget.model()._objs
        scans[0].stitching_metadata = StitchingMetadata(tomo_obj=scans[0])
        scans[1].stitching_metadata = QStitchingMetadata(tomo_obj=scans[1])

        scans[0].stitching_metadata._pos_as_m = (0.1, 1, 2)
        scans[1].stitching_metadata._pos_as_m = (-0.03, 1, 2)
        scans[2].stitching_metadata._pos_as_m = (0.23, 1, 2)

        self._widget.model().reorder_objs()
        self._widget.model().layoutChanged.emit()

        assert (
            self._widget.model().data(
                self._widget.model().index(0, 1), qt.Qt.DisplayRole
            )
            == scans[2].get_identifier().short_description()
        )
        assert (
            self._widget.model().data(
                self._widget.model().index(1, 1), qt.Qt.DisplayRole
            )
            == scans[0].get_identifier().short_description()
        )
        assert (
            self._widget.model().data(
                self._widget.model().index(2, 1), qt.Qt.DisplayRole
            )
            == scans[1].get_identifier().short_description()
        )

    def testWithoutMetadata(self):
        """
        make sure tomo obj are still displayed even if there is no metadata at all
        """
        for scan in self.scans_without_metadata[::-1]:
            self._widget.model().addTomoObj(scan)
        # try adding some metadata to object
        scans = self._widget.model()._objs
        scans[0].stitching_metadata = StitchingMetadata(tomo_obj=scans[0])
        scans[1].stitching_metadata = QStitchingMetadata(tomo_obj=scans[1])

        scans[0].stitching_metadata._pos_as_px = (12, None, 2)
        scans[1].stitching_metadata._pos_as_px = (-12, None, None)
        scans[2].stitching_metadata._pos_as_px = (9999, None, None)

        self._widget.model().reorder_objs()
        self._widget.model().layoutChanged.emit()

        assert (
            self._widget.model().data(
                self._widget.model().index(0, 1), qt.Qt.DisplayRole
            )
            == scans[2].get_identifier().short_description()
        )
        assert (
            self._widget.model().data(
                self._widget.model().index(1, 1), qt.Qt.DisplayRole
            )
            == scans[0].get_identifier().short_description()
        )
        assert (
            self._widget.model().data(
                self._widget.model().index(2, 1), qt.Qt.DisplayRole
            )
            == scans[1].get_identifier().short_description()
        )



import gc
import os
import shutil
import tempfile

import numpy
from silx.gui import qt
from silx.gui.utils.testutils import SignalListener, TestCaseQt

from orangecontrib.tomwer.widgets.control.TomoObjSeriesOW import TomoObjSeriesOW
from tomwer.core.utils.scanutils import HDF5MockContext
from tomwer.core.volume.hdf5volume import HDF5Volume


class TestTomoObjSeriesOW(TestCaseQt):
    """
    Test the TomoObjSeriesOW orange widget
    """

    def setUp(self):
        super().setUp()
        self.window = TomoObjSeriesOW()
        self.folder = tempfile.mkdtemp()

        # connect to the series created signal
        self._listener = SignalListener()
        self.window._widget.sigSeriesSelected.connect(self._listener)

    def tearDown(self):
        self.window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self.window.close()
        self.window = None
        self._listener = None
        shutil.rmtree(self.folder)
        gc.collect()

    def test_with_volumes(self):
        """test the TomoObjSeriesOW with volumes"""
        volume_1 = HDF5Volume(
            file_path=os.path.join(self.folder, "vol_file.hdf5"),
            data_path="entry0000",
            data=numpy.linspace(0, 10, 100 * 100 * 3).reshape(3, 100, 100),
        )
        volume_1.save()
        volume_2 = HDF5Volume(
            file_path=os.path.join(self.folder, "vol_file2.hdf5"),
            data_path="entry0001",
            data=numpy.linspace(60, 120, 100 * 100 * 5).reshape(5, 100, 100),
        )
        volume_2.save()

        assert self._listener.callCount() == 0
        assert len(self.window._widget._widget.getSelectedSeries()) == 0
        # add the volume to the possible objects to be added
        self.window.addTomoObj(volume_1)
        self.window.addTomoObj(volume_2)
        assert len(self.window._widget._widget.getSelectedSeries()) == 0

        # add volume to the series defined
        newSerieWidget = (
            self.window._widget._widget._seriesDefinitionWidget._manualDefWidget._newSeriesWidget
        )
        newSerieWidget.addTomoObjToCurrentSeries(volume_1)
        assert len(self.window._widget._widget.getSelectedSeries()) == 1
        newSerieWidget.addTomoObjToCurrentSeries(
            volume_1
        )  # try adding twice the same object
        assert len(self.window._widget._widget.getSelectedSeries()) == 1
        newSerieWidget.addTomoObjToCurrentSeries(volume_2)
        assert len(self.window._widget._widget.getSelectedSeries()) == 2
        assert self._listener.callCount() == 0
        self.window._widget._selectButton.released.emit()
        self.qapp.processEvents()
        assert self._listener.callCount() == 1

    def test_with_scans(self):
        """test the TomoObjSeriesOW with scans"""
        with HDF5MockContext(
            scan_path=os.path.join(self.folder, "test", "scan"), n_proj=100
        ) as scan:
            assert self._listener.callCount() == 0

            # add the scan to possible objects to be added
            self.window.addTomoObj(scan)
            assert len(self.window._widget._widget.getSelectedSeries()) == 0

            # add scan to the series defined
            newSerieWidget = (
                self.window._widget._widget._seriesDefinitionWidget._manualDefWidget._newSeriesWidget
            )
            newSerieWidget.addTomoObjToCurrentSeries(scan)

            assert len(self.window._widget._widget.getSelectedSeries()) == 1
            self.window._widget._selectButton.released.emit()
            self.qapp.processEvents()
            assert self._listener.callCount() == 1



# coding: utf-8
from __future__ import annotations

import copy
import functools
import logging
from typing import Iterable
from contextlib import AbstractContextManager

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output
from silx.gui import qt
from tomoscan.identifier import VolumeIdentifier

import tomwer.core.process.reconstruction.nabu.nabuvolume
from orangecontrib.tomwer.widgets.utils import WidgetLongProcessing
from tomwer.core import settings
from tomwer.core.cluster import SlurmClusterConfiguration
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.process.drac.processeddataset import (
    DracReconstructedVolumeDataset,
)
from tomwer.core.process.reconstruction.nabu import utils as nabu_utils
from tomwer.core.process.reconstruction.nabu.nabuvolume import NabuVolumeTask
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.char import BETA_CHAR, DELTA_CHAR
from tomwer.core.utils.scanutils import format_output_location
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.volume.volumefactory import VolumeFactory
from tomwer.gui.reconstruction.nabu.volume import NabuVolumeWindow
from tomwer.synctools.stacks.reconstruction.nabu import NabuVolumeProcessStack
from tomwer.utils import docstring

from ...orange.managedprocess import SuperviseOW

_logger = logging.getLogger(__name__)


class NabuVolumeOW(WidgetLongProcessing, SuperviseOW):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    # note of this widget should be the one registered on the documentation
    name = "nabu volume reconstruction"
    id = "orange.widgets.tomwer.reconstruction.NabuVolumeOW.NabuVolumeOW"
    description = (
        "This widget will call nabu for running a reconstruction " "on a volume"
    )
    icon = "icons/nabu_3d.svg"
    priority = 15
    keywords = ["tomography", "nabu", "reconstruction", "volume"]

    ewokstaskclass = tomwer.core.process.reconstruction.nabu.nabuvolume.NabuVolumeTask

    want_main_area = True
    resizing_enabled = True

    _ewoks_default_inputs = Setting(
        {"data": None, "nabu_volume_params": None, "nabu_params": None}
    )

    sigScanReady = qt.Signal(TomwerScanBase)
    "Signal emitted when a scan is ended"

    TIMEOUT = 30

    class Inputs:
        data = Input(
            name="data",
            type=TomwerScanBase,
            doc="one scan to be process",
            default=True,
            multiple=False,
        )
        cluster_in = Input(
            name="cluster_config",
            type=SlurmClusterConfiguration,
            doc="slurm cluster to be used",
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

        future_out = Output(
            name="future_tomo_obj",
            type=FutureTomwerObject,
            doc="data with some remote processing",
        )
        volume = Output(
            name="volume",
            type=TomwerVolumeBase,
            doc="volume(s) created",
        )
        volume_urls = Output(
            name="volume urls", type=tuple, doc="url of the volume(s) reconstructed"
        )
        data_portal_processed_datasets = Output(
            name="data_portal_processed_datasets",
            type=tuple,
            doc="data portal processed data to be saved",
        )

    class DialogCM(AbstractContextManager):
        """Simple context manager to hide / show button dialogs"""

        def __init__(self, dialogButtonsBox):
            self._dialogButtonsBox = dialogButtonsBox

        def __enter__(self):
            self._dialogButtonsBox.show()

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._dialogButtonsBox.hide()

    def __init__(self, parent=None, *args, **kwargs):
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        self._slurmCluster = None
        # processing tool
        self._processingStack = NabuVolumeProcessStack(self, process_id=self.process_id)

        _layout = gui.vBox(self.mainArea, self.name).layout()
        # main widget
        self._nabuWidget = NabuVolumeWindow(parent=self)
        _layout.addWidget(self._nabuWidget)
        # add button to validate when change reconstruction parameters is called
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        _layout.addWidget(self._buttons)

        # set up
        self._buttons.hide()

        # load settings
        nabu_volume_params = self._ewoks_default_inputs.get("nabu_volume_params", None)

        if nabu_volume_params not in (dict(), None):
            try:
                self.setConfiguration(nabu_volume_params)
            except Exception:
                _logger.warning("fail to load reconstruction settings")

        # connect signal / slot
        self._processingStack.sigComputationStarted.connect(self._startProcessing)
        self._processingStack.sigComputationEnded.connect(self._endProcessing)
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self.accept)
        self._nabuWidget.sigConfigChanged.connect(self._updateSettingsVals)

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("static_input", None)
        if static_input not in (None, {}):
            _logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    @Inputs.data
    def process(self, scan: TomwerScanBase):
        assert isinstance(scan, (TomwerScanBase, type(None)))
        if scan is None:
            return
        scan_ = copy.copy(scan)
        scan_.clear_latest_vol_reconstructions()

        # insure we are able to reconstruct
        if scan.nabu_recons_params in ({}, None):
            _logger.error(
                f"No reconstruction parameters found from nabu slices for {scan}. "
                "You should first run slice reconstruction prior to volume reconstruction"
            )
            self.Outputs.data.send(scan)
            self.sigScanReady.emit(scan)
            return

        config = scan.nabu_recons_params
        # update output format if requested
        nabuSettingsWidget = self._nabuWidget._mainWidget._nabuSettingsWidget
        if nabuSettingsWidget.redefineNabuFileFormat():
            config["output"]["file_format"] = nabuSettingsWidget.getNabuFileFormat()
        # update output location if requested
        if nabuSettingsWidget.redefineOutputLocation():
            location = nabuSettingsWidget.getNabuOutputLocation()
            if location is not None:
                location = format_output_location(location=location, scan=scan)
            config["output"]["location"] = location
            config["output"][
                "output_dir_mode"
            ] = nabuSettingsWidget._outputLocationWidget.getOutputDirMode().value

        if "phase" in config and "delta_beta" in config["phase"]:
            pag_dbs = config["phase"]["delta_beta"]
            if isinstance(pag_dbs, str):
                try:
                    pag_dbs = nabu_utils.retrieve_lst_of_value_from_str(
                        config["phase"]["delta_beta"], type_=float
                    )
                except Exception:
                    pass
            if len(pag_dbs) > 1:
                _logger.warning(
                    f"Several value found for {DELTA_CHAR} / {BETA_CHAR}. Volume reconstruction take one at most."
                )
                timeout = NabuVolumeOW.TIMEOUT if settings.isOnLbsram(scan) else None
                self._dialogDB = _DeltaBetaSelectorDialog(
                    values=pag_dbs, parent=None, timeout=timeout
                )
                self._dialogDB.setModal(False)
                self._callbackDB = functools.partial(
                    self._updateDB, scan, self._dialogDB
                )
                self._dialogDB.accepted.connect(self._callbackDB)
                self._callbackTimeout = functools.partial(self._skipProcessing, scan)
                self._dialogDB.timeoutReached.connect(self._callbackTimeout)
                self._dialogDB.show()
                return

        self._processingStack.add(scan_, self.getConfiguration())

    @docstring(SuperviseOW)
    def reprocess(self, dataset):
        if (
            dataset.axis_params is None
            or dataset.axis_params.relative_cor_value is None
        ):
            # try to retrieve last computed cor value from nabu process
            r_cor = NabuVolumeTask.retrieve_last_relative_cor(dataset)
            if dataset.axis_params is None:
                from tomwer.synctools.axis import QAxisRP

                dataset.axis_params = QAxisRP()
            try:
                dataset.axis_params.set_relative_value(float(r_cor))
            except Exception:
                pass

        self.process(dataset)

    def cancel(self, scan):
        if scan is None:
            return
        if scan in self._processingStack:
            self._processingStack.remove(scan)
        if (
            self._processingStack._data_currently_computed.get_identifier().to_str()
            == scan.get_identifier().to_str()
        ):
            # stop current processing
            self._processingStack.cancel()
            # if possible process next
            if self._processingStack.can_process_next():
                self._processingStack._process_next()

    @Inputs.cluster_in
    def setCluster(self, cluster):
        self._slurmCluster = cluster

    def _updateDB(self, scan, dialog):
        db = dialog.getSelectedValue()
        if db is not None:
            try:
                scan.nabu_recons_params["phase"]["delta_beta"] = (db,)
            except Exception as e:
                logging.error(e)
            else:
                self.process(scan=scan)

    def _endProcessing(self, scan, future_tomo_obj):
        WidgetLongProcessing._endProcessing(self, scan)
        if scan is not None:
            # send scan
            self.Outputs.data.send(scan)
            self.sigScanReady.emit(scan)

            # send volume urls
            volume_urls = []
            for volume_id in scan.latest_vol_reconstructions:
                assert isinstance(volume_id, VolumeIdentifier)
                volume_urls.extend(VolumeFactory.from_identifier_to_vol_urls(volume_id))
            if len(volume_urls) > 0:
                self.Outputs.volume_urls.send(tuple(volume_urls))

            # send volume identifier(s) and associated IcatDataBase objects
            n_rec_volumes = len(scan.latest_vol_reconstructions)
            drac_processed_datasets = []
            if n_rec_volumes > 0:
                if n_rec_volumes > 1:
                    _logger.warning(
                        f"{n_rec_volumes} volume reconstructed when at most one expected"
                    )
                try:
                    volume = VolumeFactory.create_tomo_object_from_identifier(
                        scan.latest_vol_reconstructions[0]
                    )
                except Exception as e:
                    _logger.error(
                        f"Failed to retrieve volume from {volume_id}. Error is {e}"
                    )
                else:
                    self.Outputs.volume.send(volume)

                    icatReconstructedDataset = DracReconstructedVolumeDataset(
                        tomo_obj=volume,
                        source_scan=scan,
                    )
                    drac_processed_datasets.append(icatReconstructedDataset)

            if len(drac_processed_datasets) > 0:
                self.Outputs.data_portal_processed_datasets.send(
                    drac_processed_datasets
                )

        if future_tomo_obj is not None:
            self.Outputs.future_out.send(future_tomo_obj)

    def setDryRun(self, dry_run):
        self._processingStack.setDryRun(dry_run)

    def _ciExec(self):
        self.activateWindow()
        self.raise_()
        self.show()

    def _updateSettingsVals(self):
        self._ewoks_default_inputs = {
            "data": None,
            "nabu_volume_params": self.getConfiguration(),
            "nabu_params": None,
        }

    def _skipProcessing(self, scan):
        self.Outputs.data.send(scan)
        self.sigScanReady.emit(scan)

    def getConfiguration(self):
        config = self._nabuWidget.getConfiguration()
        config["cluster_config"] = self._slurmCluster
        return config

    def setConfiguration(self, config):
        # ignore slurm cluster. Defined by the upper widget
        config.pop("cluster_config", None)
        self._nabuWidget.setConfiguration(config=config)


class _DeltaBetaSelectorDialog(qt.QDialog):
    timeoutReached = qt.Signal()

    def __init__(self, values, parent=None, timeout=None):
        """

        :param values:
        :param parent:
        :param timeout: if a timeout is provided once reach this will
                        automatically reject the delta / beta selection.
                        This is needed when on lbsram to avoid 'locking'
                        a reconstruction. In sec.
        """
        qt.QDialog.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        self.mainWidget = _DeltaBetaSelector(parent=self, values=values)
        self.layout().addWidget(self.mainWidget)

        types = qt.QDialogButtonBox.Ok | qt.QDialogButtonBox.Cancel
        self._buttons = qt.QDialogButtonBox(parent=self)
        self._buttons.setStandardButtons(types)
        self.layout().addWidget(self._buttons)
        self._timeout = timeout
        if self._timeout is None:
            self.setWindowTitle(f"Select one value for {DELTA_CHAR} / {BETA_CHAR}")
        else:
            self.setWindowTitle(
                f"Select one value for {DELTA_CHAR} / {BETA_CHAR}. (close automatically in {self._timeout} sec.)"
            )

        # connect signal / slot
        self._buttons.accepted.connect(self.accept)
        self._buttons.rejected.connect(self.reject)

        # expose API
        self.getSelectedValue = self.mainWidget.getSelectedValue

        # add timers
        if timeout is not None:
            self._timer = qt.QTimer()
            self._timer.timeout.connect(self.reject)
            self._timer.start(timeout * 1000)
            self._displayTimer = qt.QTimer()
            self._displayTimer.timeout.connect(self._updateTitle)
            self._displayTimer.start(1000)
        else:
            self._timer = None

    def reject(self):
        if self._timer:
            self._timer.stop()
            self._timer.timeout.disconnect(self.reject)
            self._displayTimer.stop()
            self._displayTimer.timeout.disconnect(self._updateTitle)
            self._timer = None

        qt.QDialog.reject(self)

    def _updateTitle(self):
        self._timeout = self._timeout - 1
        if self._timeout <= 0:
            self.timeoutReached.emit()
            self.reject()
        else:
            self.setWindowTitle(
                f"Select one value for {DELTA_CHAR} / {BETA_CHAR}. (close automatically in {self._timeout} sec.)"
            )
            self._displayTimer.start(1000)


class _DeltaBetaSelector(qt.QTableWidget):
    """Widget used to select a value of delta beta if several provided"""

    def __init__(self, values, parent=None):
        qt.QTableWidget.__init__(self, parent)
        self.setSelectionMode(qt.QAbstractItemView.SingleSelection)
        self.setHorizontalHeaderLabels([DELTA_CHAR + " / " + BETA_CHAR])
        self.setRowCount(0)
        self.setColumnCount(1)
        self.verticalHeader().hide()
        if hasattr(self.horizontalHeader(), "setSectionResizeMode"):  # Qt5
            self.horizontalHeader().setSectionResizeMode(0, qt.QHeaderView.Stretch)
        else:  # Qt4
            self.horizontalHeader().setResizeMode(0, qt.QHeaderView.Stretch)
        self.setAcceptDrops(False)

        # set up
        self.setValues(values=values)

    def setValues(self, values: Iterable):
        self.setHorizontalHeaderLabels([DELTA_CHAR + " / " + BETA_CHAR])
        self.setRowCount(len(values))
        self.setColumnCount(1)
        for i_value, value in enumerate(values):
            _item = qt.QTableWidgetItem()
            _item.setText(str(value))
            _item.setFlags(qt.Qt.ItemIsEnabled | qt.Qt.ItemIsSelectable)
            self.setItem(i_value, 0, _item)
            _item.setSelected(i_value == 0)

    def getSelectedValue(self):
        sel = None
        for item in self.selectedItems():
            sel = item.text()
        return sel



"""contains QLFileSystem class. Specific implementation of the QLineEdit to select a file path.
"""

from silx.gui import qt


class QLFileSystem(qt.QLineEdit):
    """
    QLineEdit with a completer using a QDirModel
    """

    def __init__(self, text, parent, filters=None, **kwargs):
        super().__init__(parent=parent, **kwargs)
        self.completer = qt.QCompleter()
        model = qt.QDirModel(self.completer)
        if filters is not None:
            model.setFilter(filters)
        self.completer.setModel(model)
        self.setCompleter(self.completer)
        if text is not None:
            self.setText(text)



NAME = "data portal"

ID = "orangecontrib.tomwer.widgets.dataportal"

DESCRIPTION = "widgets for data portal publication"

LONG_DESCRIPTION = "Contains widget to publish data to the esrf data portal"

ICON = "../../widgets/icons/tomwer_data_portal.png"

BACKGROUND = "#C0CCFF"

PRIORITY = 13



from __future__ import annotations

import functools
from silx.gui import qt
from tomwer.gui.utils.qt_utils import block_signals


class StepSizeSelectorWidget(qt.QGroupBox):
    """
    Widget to define some steps size (as float). Used by the Axis and the AxisOrdered widgets

    :param title: title to provide to the group box
    :param label: text for the label set at the left of the QLineEdit
    :param fine_value: (optional) value to provide for 'fine' step
    :param medium_value: (optional) value to provide for 'medium' step
    :param rough_value: (optional) value to provide for 'rough' step
    :param dtype: type of the step. Can be int or float
    """

    valueChanged = qt.Signal()
    """emit when the step size change"""

    def __init__(
        self,
        parent=None,
        title="",
        label: str = "step size",
        fine_value: float | int | None = 0.1,
        medium_value: float | int | None = 0.5,
        rough_value: float | int | None = 1.0,
        unit: str | None = "px",
        dtype: int | float = float,
    ):
        assert fine_value is None or isinstance(
            fine_value, dtype
        ), f"fine_value is expected to be None or a {dtype}. Get {type(fine_value)}"
        assert medium_value is None or isinstance(
            medium_value, dtype
        ), f"medium_value is expected to be None or a {dtype}. Get {type(medium_value)}"
        assert rough_value is None or isinstance(
            rough_value, dtype
        ), f"rough_value is expected to be None or a {dtype}. Get {type(rough_value)}"
        super().__init__(title, parent)
        self._dtype = dtype
        if unit is None:
            unit = ""
        else:
            unit = f" {unit}"
        self.setLayout(qt.QGridLayout())
        self.layout().setSpacing(2)

        # label
        self.layout().addWidget(qt.QLabel(label), 0, 0, 3, 1)

        # QLE manual step size
        default_value = medium_value or fine_value or rough_value
        self._manualLE = qt.QLineEdit(str(default_value), parent=self)
        if dtype is float:
            validator = qt.QDoubleValidator(parent=self._manualLE, decimals=2)
            validator.setBottom(0.0)
        elif dtype is int:
            validator = qt.QIntValidator(parent=self._manualLE)
            validator.setBottom(0)
        else:
            raise ValueError("dtype is expected to be int or float")
        self._manualLE.setValidator(validator)
        self.layout().addWidget(self._manualLE, 0, 1, 3, 1)

        # buttons
        buttons_font = self.font()
        buttons_font.setPixelSize(10)
        self._expectedValues = {}
        # for each button associate the expected value

        # fine
        if fine_value is not None:
            self._fineButton = qt.QPushButton(f"fine ({fine_value}{unit})", parent=self)
            self._fineButton.setCheckable(True)
            self._fineButton.setFont(buttons_font)
            self.layout().addWidget(self._fineButton, 0, 2, 1, 1)
            self._fineButton.released.connect(
                functools.partial(self.setStepSize, fine_value)
            )
            self._expectedValues[self._fineButton] = fine_value
        else:
            self._fineButton = None

        # medium
        if medium_value is not None:
            self._mediumButton = qt.QPushButton(
                f"medium ({medium_value}{unit})", parent=self
            )
            self._mediumButton.setCheckable(True)
            self._mediumButton.setFont(buttons_font)
            self.layout().addWidget(self._mediumButton, 1, 2, 1, 1)
            self._mediumButton.released.connect(
                functools.partial(self.setStepSize, medium_value)
            )
            self._expectedValues[self._mediumButton] = medium_value
        else:
            self._mediumButton = None

        # rough
        if rough_value is not None:
            self._roughButton = qt.QPushButton(
                f"rough ({rough_value}{unit})", parent=self
            )
            self._roughButton.setCheckable(True)
            self._roughButton.setFont(buttons_font)
            self.layout().addWidget(self._roughButton, 2, 2, 1, 1)
            self._roughButton.released.connect(
                functools.partial(self.setStepSize, rough_value)
            )
            self._expectedValues[self._roughButton] = rough_value
        else:
            self._roughButton = None

        # connect signal / slot
        self._manualLE.textChanged.connect(self._updateButtonChecked)
        self._manualLE.textChanged.connect(self.valueChanged)
        # set up
        self._updateButtonChecked(self._manualLE.text())

    def _updateButtonChecked(self, text):
        if text == "":
            return
        current_value = self._dtype(text)
        for button, activation_value in self._expectedValues.items():
            with block_signals(button):
                button.setChecked(activation_value == current_value)

    def getStepSize(self) -> float | int:
        """

        :return: displacement shift defined. Output is of same type as the 'dtype' provided during construction.
        """
        return self._dtype(self._manualLE.text())

    def setStepSize(self, value: float | int):
        """

        :param value: shift step
        """
        assert isinstance(value, self._dtype)
        self._manualLE.setText(str(value))



import platform
import os
import string

from processview.core.manager import ProcessManager

try:
    from ewoksnotify.tasks.email import EmailTask
except ImportError:
    has_ewoksnotify = False
else:
    has_ewoksnotify = True

from tomwer.core.process.task import Task
from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.core.volume.hdf5volume import HDF5Volume

from datetime import datetime
from tomwer.version import version as __version


class TomoEmailTask(
    Task,
    input_names=("tomo_obj", "configuration"),
    output_names=("tomo_obj",),
):
    """Dedicated task for tomography and gui approach"""

    def __init__(
        self, varinfo=None, inputs=None, node_id=None, node_attrs=None, execinfo=None
    ):
        super().__init__(varinfo, inputs, node_id, node_attrs, execinfo)
        if not has_ewoksnotify:
            raise ImportError(
                "ewoksnotify not install but required. Please run 'pip install ewoksnotify[full]'"
            )

    def run(self):
        tomo_obj = self.inputs.tomo_obj
        configuration = self.inputs.configuration
        configuration["text"] = format_email_info(
            configuration.get("text", ""), tomo_obj=tomo_obj
        )
        configuration["subject"] = format_email_info(
            configuration.get("subject", ""), tomo_obj=tomo_obj
        )
        task = EmailTask(inputs=configuration)
        task.run()
        self.outputs.tomo_obj = self.inputs.tomo_obj


def _ls_tomo_obj(tomo_obj) -> tuple:
    """
    list information regarding a tomo obj
    """
    if isinstance(tomo_obj, TomwerScanBase):
        # for tomoscan base use the `path` attribut
        file_path_to_list = tomo_obj.path
    elif isinstance(tomo_obj, TomwerVolumeBase):
        if isinstance(tomo_obj, HDF5Volume):
            file_path_to_list = os.path.dirname(tomo_obj.url.file_path())
        else:
            file_path_to_list = tomo_obj.url.file_path()
    else:
        raise TypeError

    def get_size(file_path, decimal_places=2) -> str:
        size = os.path.getsize(file_path)
        for unit in ("B", "KiB", "MiB", "GiB", "TiB", "PiB"):
            if size < 1024.0 or unit == "PiB":
                break
            size /= 1024.0
        return f"{size:.{decimal_places}f} {unit}"

    return tuple(
        [
            f"{get_size(os.path.join(file_path_to_list, file_path))} - {file_path}"
            for file_path in os.listdir(file_path_to_list)
        ]
    )


def _ls_dataset_state(tomo_obj) -> str:
    """
    list the status of all met processes by the tomo_obj
    """
    states = {}
    for process in ProcessManager().get_processes():
        state = ProcessManager().get_dataset_state(
            dataset_id=tomo_obj.get_identifier(),
            process=process,
        )
        if state is not None:
            states[process] = state

    return "\n".join(
        [f"* {process.name}: {state.value}" for process, state in states.items()]
    )


def format_email_info(my_str: str, tomo_obj: TomwerObject) -> str:
    """
    format `my_str` string. It can contain one of the following keyword:

    - {tomo_obj_short_id}: tomo_obj 'short id' (calling identifier.short_description)
    - {tomo_obj_id}: tomo_obj id
    - {ls_tomo_obj}: ls of the scan folder
    - {timestamp}: current time
    - {footnote}: some footnote defined by tomwer
    - {dataset_processing_states}: list the status of all met processing
    """

    keywords = {
        "tomo_obj_short_id": tomo_obj.get_identifier().short_description(),
        "tomo_obj_id": tomo_obj.get_identifier().to_str(),
        "ls_tomo_obj": "\n".join(_ls_tomo_obj(tomo_obj)),
        "timestamp": datetime.utcnow().isoformat(timespec="seconds"),
        "footnote": f"email send by tomwer - {__version} from {platform.node()}",
        "dataset_processing_states": _ls_dataset_state(tomo_obj),
    }

    # filter necessary keywords
    def get_necessary_keywords():
        formatter = string.Formatter()
        return [field for _, field, _, _ in formatter.parse(my_str) if field]

    requested_keywords = get_necessary_keywords()

    def keyword_needed(pair):
        keyword, _ = pair
        return keyword in requested_keywords

    keywords = dict(filter(keyword_needed, keywords.items()))
    return my_str.format(**keywords)



"""Test suite to scripts"""

import logging
import subprocess
import sys
import unittest
import os

_logger = logging.getLogger(__name__)


class TestScriptsHelp(unittest.TestCase):
    def executeCommandLine(self, command_line, env):
        """Execute a command line.

        Log output as debug in case of bad return code.
        """
        _logger.info("Execute: %s", " ".join(command_line))
        p = subprocess.Popen(
            command_line, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env=env
        )
        out, err = p.communicate()
        _logger.info("Return code: %d", p.returncode)
        try:
            out = out.decode("utf-8")
        except UnicodeError:
            pass
        try:
            err = err.decode("utf-8")
        except UnicodeError:
            pass

        if p.returncode != 0:
            _logger.info("stdout:")
            _logger.info("%s", out)
            _logger.info("stderr:")
            _logger.info("%s", err)
        else:
            _logger.debug("stdout:")
            _logger.debug("%s", out)
            _logger.debug("stderr:")
            _logger.debug("%s", err)
        self.assertEqual(p.returncode, 0)

    def executeAppHelp(self, script_name, module_name):
        test_script = ScriptTest()
        script = test_script.script_path(script_name, module_name)
        env = test_script.get_test_env()
        if script.endswith(".exe"):
            command_line = [script]
        else:
            command_line = [sys.executable, script]
        command_line.append("--help")
        self.executeCommandLine(command_line, env)

    def testAxis(self):
        self.executeAppHelp("axis", "tomwer.app.axis")

    def testCanvas(self):
        self.executeAppHelp("axis", "tomwer.app.canvas")

    def testDarkRef(self):
        self.executeAppHelp("darkref", "tomwer.app.darkref")

    def testDarkRefPath(self):
        self.executeAppHelp("darkrefpatch", "tomwer.app.darkrefpatch")

    def testDiffFrame(self):
        self.executeAppHelp("diffFrame", "tomwer.app.diffframe")

    def testImageKeyEditor(self):
        self.executeAppHelp("image-key-editor", "tomwer.app.imagekeyeditor")

    def testImageKeyUpgrader(self):
        self.executeAppHelp("image-key-upgrader", "tomwer.app.imagekeyupgrader")

    def testIntensityNormalization(self):
        self.executeAppHelp(
            "intensity-normalization", "tomwer.app.intensitynormalization"
        )

    def testMulticor(self):
        self.executeAppHelp("multicor", "tomwer.app.multicor")

    def testMultiPag(self):
        self.executeAppHelp("multipag", "tomwer.app.multipag")

    def testNabu(self):
        self.executeAppHelp("nabu", "tomwer.app.nabuapp")

    def testNxTomoEditor(self):
        self.executeAppHelp("nxtomo-editor", "tomwer.app.nxtomoeditor")

    def testPatchRawDarkflat(self):
        self.executeAppHelp("patch-raw-dark-flat", "tomwer.app.patchrawdarkflat")

    def testRadioStack(self):
        self.executeAppHelp("radiostack", "tomwer.app.radiostack")

    def testReduceDarkFlat(self):
        self.executeAppHelp("reduce-dark-flat", "tomwer.app.reducedarkflat")

    def testRSync(self):
        self.executeAppHelp("rsync", "tomwer.app.rsync")

    def testSampleMoved(self):
        self.executeAppHelp("samplemoved", "tomwer.app.samplemoved")

    def testScanviewer(self):
        self.executeAppHelp("scanviewer", "tomwer.app.scanviewer")

    def testSinogramViewer(self):
        self.executeAppHelp("sinogramviewer", "tomwer.app.sinogramviewer")

    def testSliceStack(self):
        self.executeAppHelp("slicestack", "tomwer.app.slicestack")

    def testStopDataListener(self):
        self.executeAppHelp("stop-data-listener", "tomwer.app.stopdatalistener")

    def testZStitching(self):
        self.executeAppHelp("zstitching", "tomwer.app.zstitching")


class ScriptTest:
    """
    Class providing useful stuff for preparing script tests.
    """

    def get_test_env(self):
        """
        Returns an associated environment with a working project.
        """
        env = dict((str(k), str(v)) for k, v in os.environ.items())
        env["PYTHONPATH"] = os.pathsep.join(sys.path)
        return env

    def script_path(self, script_name, module_name):
        """Returns the script path according to it's location"""
        import importlib

        module = importlib.import_module(module_name)
        script = module.__file__
        return script

    def get_installed_script_path(self, script):
        """
        Returns the path of the executable and the associated environment

        In Windows, it checks availability of script using .py .bat, and .exe
        file extensions.
        """
        paths = os.environ.get("PATH", "").split(os.pathsep)
        for base in paths:
            # clean up extra quotes from paths
            if base.startswith('"') and base.endswith('"'):
                base = base[1:-1]
        # script not found
        _logger.warning("Script '%s' not found in paths: %s", script, ":".join(paths))
        script_path = script
        return script_path



# coding: utf-8


from ewokscore.task import Task as EwoksTask


class _ImageStackViewerPlaceHolder(EwoksTask, input_names=("data",)):
    """
    Task to aggregate several images
    """

    def run(self):
        pass



# coding: utf-8


from ewokscore.task import Task as EwoksTask


class _RadioStackPlaceHolder(EwoksTask, input_names=("data",)):
    """
    Task to aggregate radios of several scans and allow to browse along them all
    """

    def run(self):
        pass



# coding: utf-8
from __future__ import annotations


import logging

from silx.gui import qt
from silx.utils.enum import Enum as _Enum

from tomwer.core.process.reconstruction.nabu.utils import _NabuPhaseMethod, _NabuStages
from tomwer.core.utils.char import BETA_CHAR, DELTA_CHAR
from tomwer.gui.reconstruction.nabu.nabuconfig import base
from tomwer.gui.reconstruction.nabu.nabuconfig.ctf import CTFConfig
from tomwer.gui.utils.inputwidget import SelectionLineEdit
from tomwer.gui.utils.scrollarea import QComboBoxIgnoreWheel as QComboBox
from tomwer.gui import icons
from tomwer.utils import docstring

_logger = logging.getLogger(__name__)


class PaddingMode(_Enum):
    ZEROS = "zeros"
    MEAN = "mean"
    EDGE = "edge"
    SYMMETRIC = "symmetric"
    REFLECT = "reflect"


class _NabuPhaseConfig(qt.QWidget, base._NabuStageConfigBase):
    """
    Widget to define the configuration of the nabu preprocessing
    """

    sigConfChanged = qt.Signal(str)
    """Signal emitted when the configuration change. Parameter is the option
    modified
    """

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent, stage=_NabuStages.PHASE)
        base._NabuStageConfigBase.__init__(self, stage=_NabuStages.PHASE)
        self.setLayout(qt.QGridLayout())

        # phase method
        self._methodLabel = qt.QLabel("method", self)
        self.layout().addWidget(self._methodLabel, 1, 0, 1, 1)
        self._methodCB = QComboBox(parent=self)
        for method in _NabuPhaseMethod:
            self._methodCB.addItem(method.value)
        idx_ctf = self._methodCB.findText(_NabuPhaseMethod.CTF.value)
        self._methodCB.setItemData(
            idx_ctf, "Contrast Transfer Function", qt.Qt.ToolTipRole
        )
        self.layout().addWidget(self._methodCB, 1, 1, 1, 3)
        self.registerWidget(self._methodLabel, "required")
        self.registerWidget(self._methodCB, "required")

        # paganin & ctf options
        self._paganinOpts = NabuPaganinConfig(parent=self)
        self.layout().addWidget(self._paganinOpts, 2, 0, 3, 3)

        # unsharp options
        self._unsharpOpts = NabuUnsharpConfig(parent=self)
        self.layout().addWidget(self._unsharpOpts, 6, 0, 3, 3)

        # ctf options
        self._ctfOpts = CTFConfig(parent=self)
        self.layout().addWidget(self._ctfOpts, 9, 0, 3, 4)
        self.registerWidget(self._ctfOpts, "advanced")

        # spacer for style
        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 200, 1, 1, 1)

        # set up
        item_index = self._methodCB.findText(_NabuPhaseMethod.PAGANIN.value)
        assert item_index >= 0
        self._methodCB.setCurrentIndex(item_index)

        # connect signal / slot
        self._methodCB.currentIndexChanged.connect(self._methodChanged)
        self._paganinOpts.sigConfChanged.connect(self.sigConfChanged)
        self._unsharpOpts.sigConfChanged.connect(self.sigConfChanged)
        self._ctfOpts.sigConfChanged.connect(self.sigConfChanged)

        # set up
        self._paganinOpts.setEnabled(self.getMethod() is not _NabuPhaseMethod.NONE)
        self._ctfOpts.setEnabled(self.getMethod() is _NabuPhaseMethod.CTF)

    def _methodChanged(self, *args, **kwargs):
        self._paganinOpts.setEnabled(self.getMethod() is not _NabuPhaseMethod.NONE)
        self._ctfOpts.setEnabled(self.getMethod() is _NabuPhaseMethod.CTF)
        self._unsharpOpts.setDiscourageUnsharpMask(
            self.getMethod() is _NabuPhaseMethod.NONE
        )
        self.sigConfChanged.emit("method")

    def _signalConfChanged(self, param):
        self.sigConfChanged.emit(param)

    def getMethod(self) -> _NabuPhaseMethod:
        return _NabuPhaseMethod.from_value(self._methodCB.currentText())

    @docstring(base._NabuStageConfigBase)
    def setConfiguration(self, config) -> None:
        if "method" in config:
            method = config.get("method", _NabuPhaseMethod.NONE)
            if method == "none":
                method = _NabuPhaseMethod.NONE
            self._paganinOpts.setConfiguration(config)
            method = _NabuPhaseMethod.from_value(method)
            index_method = self._methodCB.findText(method.value)
            if index_method >= 0:
                self._methodCB.setCurrentIndex(index_method)
            else:
                _logger.warning("unable to find method {method}")
        self._unsharpOpts.setConfiguration(config)
        self._ctfOpts.setConfiguration(config)

    @docstring(base._NabuStageConfigBase)
    def getConfiguration(self) -> dict:
        configuration = {"method": self.getMethod().value}
        if self.getMethod() in (_NabuPhaseMethod.PAGANIN, _NabuPhaseMethod.CTF):
            configuration.update(self._paganinOpts.getConfiguration())
        configuration.update(self._unsharpOpts.getConfiguration())
        configuration.update(self._ctfOpts.getConfiguration())
        return configuration

    def setConfigurationLevel(self, level):
        base._NabuStageConfigBase.setConfigurationLevel(self, level)
        self._unsharpOpts.setConfigurationLevel(level=level)
        self._paganinOpts.setConfigurationLevel(level=level)

    def setDeltaBetaValue(self, value):
        self._paganinOpts.setDeltaBetaValue(value)

    def getUnsharpCoeff(self) -> float:
        return self._unsharpOpts.getUnsharpCoeff()

    def setUnsharpCoeff(self, coeff: float):
        self._unsharpOpts.setUnsharpCoeff(coeff)

    def getUnsharpSigma(self) -> float:
        return self._unsharpOpts.getUnsharpSigma()

    def setUnsharpSigma(self, sigma: float):
        self._unsharpOpts.setUnsharpSigma(sigma)

    def getPaddingType(self):
        return self._paganinOpts.getPaddingType()

    def setPaddingType(self, padding_type):
        return self._paganinOpts.setPaddingType(padding_type)


class NabuPaganinConfig(qt.QWidget, base._NabuStageConfigBase):
    """Configuration widget dedicated to the paganin options for nabu"""

    sigConfChanged = qt.Signal(str)
    """Signal emitted when the configuration change. Parameter is the option
    modified
    """

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent=parent, stage=_NabuStages.PHASE)
        base._NabuStageConfigBase.__init__(self, stage=_NabuStages.PHASE)
        self.setLayout(qt.QGridLayout())

        # paganin delta / beta
        label = DELTA_CHAR + " / " + BETA_CHAR
        self._db_label = qt.QLabel(label, self)
        self.layout().addWidget(self._db_label, 0, 0, 1, 1)
        self._deltaBetaQLE = SelectionLineEdit(
            "100.0", self, allow_negative_indices=False
        )
        self.layout().addWidget(self._deltaBetaQLE, 0, 1, 1, 3)
        self.registerWidget(self._db_label, "required")
        self.registerWidget(self._deltaBetaQLE, "required")

        # paganin padding_type
        self._paddingLabel = qt.QLabel("padding", self)
        self.layout().addWidget(self._paddingLabel, 2, 0, 1, 1)
        self._paddingTypeCB = QComboBox(self)
        self._paddingTypeCB.setToolTip(
            "Padding type for the filtering step " "in Paganin/CTR."
        )
        for padding_type in (PaddingMode.ZEROS, PaddingMode.EDGE):
            self._paddingTypeCB.addItem(padding_type.value)
        self.layout().addWidget(self._paddingTypeCB, 2, 1, 1, 3)
        self.registerWidget(self._paddingLabel, "advanced")
        self.registerWidget(self._paddingTypeCB, "advanced")

        # set up
        item_index = self._paddingTypeCB.findText(PaddingMode.EDGE.value)
        self._paddingTypeCB.setCurrentIndex(item_index)

        # connect signal - slot
        self._deltaBetaQLE.editingFinished.connect(self._paganinDBChanged)
        self._paddingTypeCB.currentIndexChanged.connect(self._paganinPaddingTypeChanged)

    def _paganinDBChanged(self, *args, **kwargs):
        self.sigConfChanged.emit("delta_beta")

    def _paganinMargeChanged(self, *args, **kwargs):
        self.sigConfChanged.emit("marge")

    def _paganinPaddingTypeChanged(self, *args, **kwargs):
        self.sigConfChanged.emit("padding_type")

    def getPaddingType(self) -> PaddingMode:
        current_text = self._paddingTypeCB.currentText()
        return PaddingMode.from_value(current_text)

    def setPaddingType(self, padding_type):
        padding_type = PaddingMode.from_value(padding_type)
        item_index = self._paddingTypeCB.findText(padding_type.value)
        self._paddingTypeCB.setCurrentIndex(item_index)

    def getDeltaBeta(self) -> str:
        return self._deltaBetaQLE.text()

    def setDeltaBetaValue(self, value):
        self._deltaBetaQLE.setText(str(value))

    def getConfiguration(self):
        return {
            "delta_beta": self.getDeltaBeta(),  # this one is not cast because can contain several values
            "padding_type": self.getPaddingType().value,
        }

    def setConfiguration(self, conf):
        # delta_beta
        delta_beta = conf.get("delta_beta", None)
        if delta_beta is not None:
            self._deltaBetaQLE.setText(str(delta_beta))
        # padding_type
        padding_type = conf.get("padding_type", None)
        if padding_type is not None:
            self.setPaddingType(padding_type)


class NabuUnsharpConfig(qt.QWidget, base._NabuStageConfigBase):
    """Configuration widget dedicated to the unsharp options for nabu"""

    sigConfChanged = qt.Signal(str)
    """Signal emitted when the configuration change. Parameter is the option
    modified
    """

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent=parent, stage=_NabuStages.PHASE)
        base._NabuStageConfigBase.__init__(self, stage=_NabuStages.PHASE)
        self.__discourageUnSharpMask = False
        self.setLayout(qt.QGridLayout())

        warning_icon = icons.getQIcon("warning")
        warning_unsharp_no_phase = "Warning: applying unsharp mask without phase retrieval can lead to 'noisy' reconstruction"

        # unsharp coeff
        self._unsharpCoeffCB = qt.QCheckBox("unsharp coeff", self)
        self._unsharpCoeffCB.setToolTip(
            "Unsharp mask strength. The unsharped "
            "image is equal to\n  UnsharpedImage "
            "=  (1 + coeff)*originalPaganinImage "
            "- coeff * ConvolvedImage. Setting "
            "this coefficient to zero means that "
            "no unsharp mask will be applied."
        )
        self.layout().addWidget(self._unsharpCoeffCB, 0, 0, 1, 1)
        self._unsharpCoeffQLE = qt.QLineEdit("", self)
        self._unsharpCoeffQLE.setValidator(qt.QDoubleValidator())
        self.layout().addWidget(self._unsharpCoeffQLE, 0, 1, 1, 1)
        self.registerWidget(self._unsharpCoeffCB, "optional")
        self._unsharpCoeffOpt = self.registerWidget(self._unsharpCoeffQLE, "optional")
        # unsharp coeff warning
        self._unSharpCoeffWarning = qt.QLabel("")
        self._unSharpCoeffWarning.setToolTip(warning_unsharp_no_phase)
        self._unSharpCoeffWarning.setPixmap(warning_icon.pixmap(20, state=qt.QIcon.On))
        self.layout().addWidget(self._unSharpCoeffWarning, 0, 2, 1, 1)
        self.registerWidget(self._unSharpCoeffWarning, "optional")

        # unsharp_sigma
        self._unsharpSigmaCB = qt.QCheckBox("unsharp sigma", self)
        self._unsharpSigmaCB.setToolTip(
            "Standard deviation of the Gaussian "
            "filter when applying an unsharp "
            "mask\nafter the Paganin filtering. "
            "Disabled if set to 0."
        )
        self.layout().addWidget(self._unsharpSigmaCB, 1, 0, 1, 1)
        self._unsharpSigmaQLE = qt.QLineEdit("", self)
        self._unsharpSigmaQLE.setValidator(qt.QDoubleValidator())
        self.layout().addWidget(self._unsharpSigmaQLE, 1, 1, 1, 1)
        self.registerWidget(self._unsharpSigmaCB, "optional")
        self._unsharpSigmaOpt = self.registerWidget(self._unsharpSigmaQLE, "optional")
        # unsharp coeff warning
        self._unSharpSigmaWarning = qt.QLabel("")
        self._unSharpSigmaWarning.setToolTip(warning_unsharp_no_phase)
        self._unSharpSigmaWarning.setPixmap(warning_icon.pixmap(20, state=qt.QIcon.On))
        self.layout().addWidget(self._unSharpSigmaWarning, 1, 2, 1, 1)
        self.registerWidget(self._unSharpSigmaWarning, "optional")

        # set up
        self._unsharpCoeffCB.setChecked(False)
        self._unsharpCoeffQLE.setText(str(3.0))
        self._unsharpSigmaCB.setChecked(False)
        self._unsharpSigmaQLE.setText(str(0.8))
        self._showUnsharpCoeffQLE(False)
        self._showUnsharpSigmaQLE(False)
        self._updateUnsharpMaskWarning()

        # signal / slot connection
        ## unsharp coeff
        self._unsharpCoeffCB.toggled.connect(self._showUnsharpCoeffQLE)
        self._unsharpCoeffCB.toggled.connect(self._unsharpCoeffChanged)
        self._unsharpCoeffCB.toggled.connect(self._updateUnsharpMaskWarning)
        self._unsharpCoeffQLE.editingFinished.connect(self._unsharpCoeffChanged)
        ## unsharp sigma
        self._unsharpSigmaCB.toggled.connect(self._showUnsharpSigmaQLE)
        self._unsharpSigmaCB.toggled.connect(self._unsharpSigmaChanged)
        self._unsharpSigmaCB.toggled.connect(self._updateUnsharpMaskWarning)
        self._unsharpSigmaQLE.editingFinished.connect(self._unsharpSigmaChanged)

    def _unsharpCoeffChanged(self, *args, **kwargs):
        self.sigConfChanged.emit("unsharp_coeff")

    def _unsharpSigmaChanged(self, *args, **kwargs):
        self.sigConfChanged.emit("unsharp_sigma")

    def isUnsharpCoeffActive(self):
        return self._unsharpCoeffCB.isChecked()

    def isUnsharpSigmaActive(self):
        return self._unsharpSigmaCB.isChecked()

    def _showUnsharpCoeffQLE(self, visible):
        self._unsharpCoeffOpt.setVisible(visible)

    def _showUnsharpSigmaQLE(self, visible):
        self._unsharpSigmaOpt.setVisible(visible)

    def _signalConfChanged(self, param):
        self.sigConfChanged.emit(param)

    def getUnsharpCoeff(self) -> float | int:
        if self.isUnsharpCoeffActive():
            return float(self._unsharpCoeffQLE.text())
        else:
            return 0

    def setUnsharpCoeff(self, coeff):
        if coeff == 0.0:
            self._unsharpCoeffCB.setChecked(False)
        else:
            self._unsharpCoeffCB.setChecked(True)
            self._unsharpCoeffQLE.setText(str(coeff))

    def getUnsharpSigma(self) -> float | int:
        if self.isUnsharpSigmaActive():
            return float(self._unsharpSigmaQLE.text())
        else:
            return 0

    def setUnsharpSigma(self, sigma):
        if sigma == 0.0:
            self._unsharpSigmaCB.setChecked(False)
        else:
            self._unsharpSigmaCB.setChecked(True)
            self._unsharpSigmaQLE.setText(str(sigma))

    def getConfiguration(self):
        return {
            "unsharp_coeff": self.getUnsharpCoeff(),
            "unsharp_sigma": self.getUnsharpSigma(),
        }

    def setConfiguration(self, conf):
        if "unsharp_coeff" in conf:
            self.setUnsharpCoeff(float(conf["unsharp_coeff"]))
        if "unsharp_sigma" in conf:
            self.setUnsharpSigma(float(conf["unsharp_sigma"]))

    def setDiscourageUnsharpMask(self, discourage: bool):
        """Should we discourage the user to apply unsharp mask.
        This is the case if he tries to apply an unsharp mask without phase retrieval.
        """
        self.__discourageUnSharpMask = discourage

    def _updateUnsharpMaskWarning(self):
        self._unSharpCoeffWarning.setVisible(
            self.__discourageUnSharpMask and self._unsharpCoeffCB.isChecked()
        )
        self._unSharpSigmaWarning.setVisible(
            self.__discourageUnSharpMask and self._unsharpSigmaCB.isChecked()
        )

    def setConfigurationLevel(self, level):
        super().setConfigurationLevel(level)
        # we also need to update the 'unsharp warning'
        # note: at the moment if the basic configuration level is selected, without phase and some mask
        # a warning will appear on bottom of the pahse GB. This is considered expected as this way we
        # won't hide any warning.
        self._updateUnsharpMaskWarning()



from __future__ import annotations
import os
import logging
from silx.utils.enum import Enum as _Enum
from tomwer.io.utils.raw_and_processed_data import (
    to_processed_data_path,
)
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.process.output import ProcessDataOutputDirMode
from tomwer.core.utils.scanutils import format_output_location

_logger = logging.getLogger(__name__)


PROCESS_FOLDER_RECONSTRUCTED_VOLUMES = "reconstructed_volumes"

PROCESS_FOLDER_RECONSTRUCTED_SLICES = "reconstructed_slices"


class NabuOutputFileFormat(_Enum):
    TIFF = "tiff"
    HDF5 = "hdf5"
    JP2K = "jp2"
    EDF = "edf"
    RAW = "vol"

    @classmethod
    def from_value(cls, value):
        if isinstance(value, str):
            value = value.lstrip(".")
        return super().from_value(value)


def get_file_format(file_str):
    extension = os.path.splitext(file_str.lower())[-1]
    extension = extension.lstrip(".")
    if extension in ("tiff", "tif"):
        return NabuOutputFileFormat.TIFF
    elif extension in ("hdf5", "hdf", "h5"):
        return NabuOutputFileFormat.HDF5
    elif extension in ("jp2", "jp2k", "jpg2k"):
        return NabuOutputFileFormat.JP2K
    elif extension in ("edf",):
        return NabuOutputFileFormat.EDF
    elif extension in ("vol", "raw"):
        return NabuOutputFileFormat.RAW
    else:
        raise ValueError(f"Unrecognized file extension {extension} from {file_str}")


def get_output_folder_from_scan(
    mode: ProcessDataOutputDirMode,
    scan: TomwerScanBase,
    nabu_location: str | None,
    file_basename: str,
    file_format: NabuOutputFileFormat,
    processed_data_folder_name: str | None,
) -> tuple[str, str]:
    """

    :param mode: output mode, should save this to raw data, processed data...
    :param scan: scan for which we want to get the output folder
    :param nabu_location: output location provided by user (in case mode is 'other')
    :param file_basename: file basename to take in order to create output files
    :param file_format: output volume format (edf...)
    :param processed_data_folder_name: name of the processed data folder. Like 'reconstructed_volumes' or 'reconstructed_slices'...
    :return: (location, location_cfg_files). Location is the nabu configuration field 'output/location' 'location_cfg_files' is the information on where to save the nabu configuration file

    """
    output_mode = ProcessDataOutputDirMode.from_value(mode)
    file_format = NabuOutputFileFormat.from_value(file_format)

    if output_mode is ProcessDataOutputDirMode.OTHER and nabu_location in ("", None):
        _logger.error(
            "'other' output dir requested but no path provided. Fall back on the output dir to the scan folder"
        )
        # note: this is only an info because we expect to pass by this one for all .ows configuration (before 1.3 version)
        # as there was no different option by the time
        output_mode = ProcessDataOutputDirMode.IN_SCAN_FOLDER

    if output_mode is ProcessDataOutputDirMode.OTHER:
        location = format_output_location(nabu_location, scan=scan)
        location_cfg_files = location
    elif output_mode in (
        ProcessDataOutputDirMode.IN_SCAN_FOLDER,
        ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER,
    ):
        # otherwise default location will be the data root level
        location = os.path.join(scan.path, processed_data_folder_name)
        location_cfg_files = location
        if file_format in (
            NabuOutputFileFormat.EDF.value,
            NabuOutputFileFormat.TIFF.value,
            NabuOutputFileFormat.JP2K.value,
        ):  # if user specify the location
            location = "/".join([location, file_basename])
        if output_mode is ProcessDataOutputDirMode.PROCESSED_DATA_FOLDER:
            location = to_processed_data_path(location)
            location_cfg_files = to_processed_data_path(location_cfg_files)
    else:
        raise NotImplementedError(f"mode {output_mode.value} is not implemented yet")

    return location, location_cfg_files



import logging
import os

from silx.gui import qt
from tomoscan.esrf.volume.utils import guess_volumes

from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.tomwer_object import TomwerObject
from tomwer.core.volume.edfvolume import EDFVolume, EDFVolumeIdentifier
from tomwer.core.volume.hdf5volume import HDF5Volume, HDF5VolumeIdentifier
from tomwer.core.volume.jp2kvolume import JP2KVolume, JP2KVolumeIdentifier
from tomwer.core.volume.rawvolume import RawVolume, RawVolumeIdentifier
from tomwer.core.volume.tiffvolume import (
    MultiTIFFVolume,
    MultiTiffVolumeIdentifier,
    TIFFVolume,
    TIFFVolumeIdentifier,
)
from tomwer.gui.dialog.QDataDialog import QDataDialog
from tomwer.gui.qlefilesystem import QLFileSystem

_logger = logging.getLogger(__name__)


class SingleTomoObj(qt.QWidget):
    sigTomoObjChanged = qt.Signal(str)
    """signal emit when the tomo object changed. Parameter is the identifier"""

    def __init__(self, parent=None, *args, **kwargs) -> None:
        super().__init__(parent, *args, **kwargs)
        self.setLayout(qt.QHBoxLayout())

        self._tomoObjIdentifierLineQLE = _TomoObjQLE("", self)
        self._tomoObjIdentifierLineQLE.setPlaceholderText(
            "scheme:obj_type:path?queries"
        )
        self.layout().addWidget(self._tomoObjIdentifierLineQLE)

        self._selectScanPB = qt.QPushButton("select", self)
        self.layout().addWidget(self._selectScanPB)

        self.setToolTip(
            "Define a Tomo object. You can also copy / paster folder or files to use the automatic detection. Each time it changed if fit a valid scan then this scan will be triggered"
        )

        # signal / slot connect
        self._tomoObjIdentifierLineQLE.editingFinished.connect(self._scanChanged)
        self._selectScanPB.released.connect(self._selectObject)

    def getTomoObjIdentifier(self) -> str:
        return self._tomoObjIdentifierLineQLE.text()

    def _selectObject(self, *args, **kwargs):  # pragma: no cover
        dialog = QDataDialog(self, multiSelection=True)
        dialog.setNameFilters(
            [
                "HDF5 files (*.h5 *.hdf5 *.nx *.nxs *.nexus)",
                "Nexus files (*.nx *.nxs *.nexus)",
                "Any files (*)",
            ]
        )

        if not dialog.exec_():
            dialog.close()
            return
        added_tomo_objs = []
        for path in dialog.selectedFiles():
            try:
                objs = self.guessTomoObj(path=path)
            except Exception as e:  # noqa E722
                _logger.error(f"Fail to create tomo object from {path}. Error is {e}")
            else:
                if objs is not None:
                    added_tomo_objs.extend(objs)
        if len(added_tomo_objs) > 0:
            self.setTomoObject(added_tomo_objs[0])

    def _scanChanged(self):
        obj_identifier = self.getTomoObjIdentifier()
        # if the identifier fullill a path then this is not an identifier but
        # we must find the tomo obj associated
        if os.path.exists(obj_identifier):
            path = obj_identifier
            try:
                objs = self.guessTomoObj(path=path)
            except Exception as e:  # noqa E722
                _logger.error(f"Fail to create tomo object from {path}. Error is {e}")
            else:
                if objs is not None and len(objs) > 0:
                    old = self._tomoObjIdentifierLineQLE.blockSignals(True)
                    self._tomoObjIdentifierLineQLE.setText(
                        objs[0].get_identifier().to_str()
                    )
                    self._tomoObjIdentifierLineQLE.blockSignals(old)

        self.sigTomoObjChanged.emit(self.getTomoObjIdentifier())

    def setTomoObject(self, scan):
        if isinstance(scan, TomwerObject):
            text = scan.get_identifier().to_str()
        else:
            text = scan
        self._tomoObjIdentifierLineQLE.setText(text)
        self.sigTomoObjChanged.emit(self.getTomoObjIdentifier())

    def guessTomoObj(self, path):
        path = os.path.abspath(path)

        try:
            scans = ScanFactory.create_scan_objects(path)
            if scans is None or len(scans) == 0:
                raise ValueError
        except:  # noqa E722
            try:
                volumes = guess_volumes(
                    path,
                    scheme_to_vol={
                        EDFVolumeIdentifier.scheme: EDFVolume,
                        HDF5VolumeIdentifier.scheme: HDF5Volume,
                        TIFFVolumeIdentifier.scheme: TIFFVolume,
                        MultiTiffVolumeIdentifier.scheme: MultiTIFFVolume,
                        JP2KVolumeIdentifier.scheme: JP2KVolume,
                        RawVolumeIdentifier.scheme: RawVolume,
                    },
                )
            except:  # noqa E722
                pass
                return None
            else:
                # filter potential 'nabu histogram'
                if volumes is not None:

                    def is_not_histogram(vol_identifier):
                        return not (
                            hasattr(vol_identifier, "data_path")
                            and vol_identifier.data_path.endswith("histogram")
                        )

                    volumes = tuple(filter(is_not_histogram, volumes))
                return volumes
        else:
            return scans


class _TomoObjQLE(QLFileSystem):
    """QLineEdit that try to get a Tomo object identifier once dropped"""

    def supportedDropActions(self):
        """Inherited method to redefine supported drop actions."""
        return qt.Qt.CopyAction | qt.Qt.MoveAction

    def dropEvent(self, a0) -> None:
        if hasattr(a0, "mimeData") and a0.mimeData().hasFormat("text/uri-list"):
            for url in a0.mimeData().urls():
                try:
                    new_objs = self.parent().guessTomoObj(url.path())
                except:  # noqa E722
                    pass
                else:
                    if new_objs is not None and len(new_objs) > 0:
                        self.setText(new_objs[0].get_identifier().to_str())
            a0.accept()
            self.editingFinished.emit()
        else:
            return super().dropEvent(a0)

    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            event.accept()
            event.setDropAction(qt.Qt.CopyAction)
        else:
            qt.QListWidget.dragEnterEvent(self, event)

    def dragMoveEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            event.setDropAction(qt.Qt.CopyAction)
            event.accept()
        else:
            qt.QListWidget.dragMoveEvent(self, event)
